import os
import numpy as np

from hysop import Box, Simulation, Problem, MPIParams, Field
from hysop.defaults import VelocityField, VorticityField, \
                           EnstrophyParameter, TimeParameters
from hysop.parameters.tensor_parameter import TensorParameter
from hysop.constants import Implementation, AdvectionCriteria, HYSOP_REAL, \
    StretchingFormulation, StretchingCriteria
from hysop.operators import Advection, StaticDirectionalStretching, Diffusion, \
                            PoissonCurl, AdaptiveTimeStep,                  \
                            Enstrophy, MinMaxFieldStatistics, StrangSplitting,    \
                            ParameterPlotter, PenalizeVorticity, FlowRateCorrection, \
                            VorticityAbsorption, CustomOperator
from hysop.numerics.odesolvers.runge_kutta import RK2
from hysop.methods import SpaceDiscretization, Remesh, TimeIntegrator, \
                          ComputeGranularity, Interpolation, StrangOrder
from hysop.topology.cartesian_topology import CartesianTopology
from hysop.tools.parameters import CartesianDiscretization


pi  = np.pi
cos = np.cos
sin = np.sin


# Define the domain
dim = 3
npts = (32,32,64)
box = Box(dim=dim,
          origin=[-2.56, -2.56, -2.56],
          length=[5.12, 5.12, 10.24])
cfl = 0.5
lcfl = 0.125
uinf = 1.0
viscosity = 1. / 250.
outfreq = 10
dt0 = 0.0125

# Get default MPI Parameters from domain (even for serial jobs)
mpi_params = MPIParams(comm=box.task_comm,
                       task_id=box.current_task())

# Setup usual implementation specific variables
impl = None
extra_op_kwds = {'mpi_params': mpi_params}


# ====== Sphere inside the domain ======
RADIUS = 0.5
pos = [0., 0., 0.]
def computeSphere(data, coords):
    (x, y, z) = coords[0]
    dx = x[0,0,1] - x[0,0,0]
    dy = y[0,1,0] - y[0,0,0]
    dz = z[1,0,0] - z[0,0,0]
    data[0][...] = 0.
    chi = lambda x,y,z: np.sqrt((x-pos[0])*(x-pos[0])+(y-pos[1])*(y-pos[1])+(z-pos[2])*(z-pos[2]))<=RADIUS
    data[0][chi(x,y,z)] = 1.

    # Smooth the sphere surface with a Volume-of-fluid fraction
    vof = 5  # number of points in the subgrid
    front_z = np.where(np.abs(data[0][0:-1,:,:]-data[0][1:,:,:])>0.1)
    front_z = (np.concatenate((front_z[0],front_z[0]+1)),
               np.concatenate((front_z[1],front_z[1])),
               np.concatenate((front_z[2],front_z[2])))
    front_y = np.where(np.abs(data[0][:,0:-1,:]-data[0][:,1:,:])>0.1)
    front_y = (np.concatenate((front_y[0],front_y[0])),
               np.concatenate((front_y[1],front_y[1]+1)),
               np.concatenate((front_y[2],front_y[2])))
    front = (np.concatenate((front_z[0],front_y[0])),
             np.concatenate((front_z[1],front_y[1])),
             np.concatenate((front_z[2],front_y[2])))
    front_x = np.where(np.abs(data[0][:,:,0:-1]-data[0][:,:,1:])>0.1)
    front_x = (np.concatenate((front_x[0],front_x[0])),
               np.concatenate((front_x[1],front_x[1])),
               np.concatenate((front_x[2],front_x[2]+1)))
    front = (np.concatenate((front[0],front_x[0])),
             np.concatenate((front[1],front_x[1])),
             np.concatenate((front[2],front_x[2])))

    for k,j,i in zip(*front):
        sx = np.linspace(x[0,0,i]-dx/2,x[0,0,i]+dx/2,vof)[np.newaxis,np.newaxis,:]
        sy = np.linspace(y[0,j,0]-dy/2,y[0,j,0]+dy/2,vof)[np.newaxis,:,np.newaxis]
        sz = np.linspace(z[k,0,0]-dz/2,z[k,0,0]+dz/2,vof)[:,np.newaxis,np.newaxis]
        data[0][k,j,i] = 1.*(np.sum(chi(sx, sy, sz))/(1.0*vof**3))

    return data


# ======= Function to compute initial velocity  =======
def computeVel(data, coords):
    data[0][...] = uinf
    data[1][...] = 0.
    data[2][...] = 0.
    return data


# ======= Function to compute initial vorticity =======
def computeVort(data, coords):
    data[0][...] = 0.
    data[1][...] = 0.
    data[2][...] = 0.
    return data



method = {}

# Define parameters and field (time, timestep, velocity, vorticity, enstrophy)
t, dt = TimeParameters(dtype=HYSOP_REAL)
velo = VelocityField(domain=box, dtype=HYSOP_REAL)
vorti = VorticityField(velocity=velo, dtype=HYSOP_REAL)
sphere = Field(domain=box, name="Sphere", is_vector=False, dtype=HYSOP_REAL)
wdotw = Field(domain=box, dtype=HYSOP_REAL, is_vector=False, name="WdotW")
enstrophy = EnstrophyParameter(dtype=HYSOP_REAL)
flowrate = TensorParameter(name="flowrate", dtype=HYSOP_REAL, shape=(3, ),
                           initial_value=[0., 0., uinf * box.length[1] * box.length[0]])


# Topologies
topo_nogh = CartesianTopology(domain=box,
                              discretization=CartesianDiscretization(npts,
                                  default_boundaries=True),
                              mpi_params=mpi_params,
                              cutdirs=[False, False, True])
topo_gh = CartesianTopology(domain=box,
                            discretization=CartesianDiscretization(npts,
                                ghosts=(4, 4, 4), default_boundaries=True),
                            mpi_params=mpi_params,
                            cutdirs=[False, False, True])


### Build the directional operators
#> Directional advection
advec = Advection(
    implementation=Implementation.FORTRAN,
    name='advec',
    velocity=velo,
    advected_fields=(vorti,),
    variables={velo: topo_nogh, vorti: topo_nogh},
    dt=dt, **extra_op_kwds)
#> Directional stretching + diffusion
stretch = StaticDirectionalStretching(
    implementation=Implementation.PYTHON,
    name='stretch',
    formulation=StretchingFormulation.CONSERVATIVE,
    velocity=velo,
    vorticity=vorti,
    variables={velo: topo_gh, vorti: topo_gh},
    dt=dt, **extra_op_kwds)
#> Directional splitting operator subgraph
splitting = StrangSplitting(splitting_dim=dim,
                            order=StrangOrder.STRANG_FIRST_ORDER)
splitting.push_operators(stretch)
#> Penalization
penal = PenalizeVorticity(
    implementation=Implementation.PYTHON,
    name='penalization',
    velocity=velo, vorticity=vorti,
    variables={velo: topo_gh, vorti: topo_gh, sphere: topo_gh},
    obstacles=[sphere, ], coeff=1e8,
    dt=dt, **extra_op_kwds)
#> Diffusion operator
diffuse = Diffusion(
    implementation=Implementation.FORTRAN,
    name='diffuse',
    nu=viscosity,
    Fin=vorti,
    variables={vorti: topo_nogh},
    dt=dt, **extra_op_kwds)
#> Vorticity absorption
absorption = VorticityAbsorption(
    implementation=Implementation.PYTHON,
    velocity=velo, vorticity=vorti,
    start_coord=6.68,
    flowrate=flowrate,
    name="absorption",
    variables={velo: topo_nogh, vorti: topo_nogh},
    dt=dt, **extra_op_kwds)
#> Poisson operator to recover the velocity from the vorticity
poisson = PoissonCurl(
    implementation=Implementation.FORTRAN,
    name='poisson',
    velocity=velo,
    vorticity=vorti,
    variables={velo: topo_nogh, vorti: topo_nogh},
    projection=None,
    **extra_op_kwds)
#> Flowrate correction operator to adjust velocity with prescribed flowrate
def computeFlowrate(t, flowrate):
    fr = np.zeros(3)
    fr[0] = uinf * box.length[1] * box.length[0]
    Tstart=3.0
    if t() >= Tstart and t() <= Tstart + 1.0:
        fr[1] = sin(pi * (t() - Tstart)) * \
                      box.length[1] * box.length[0]
    flowrate.value = fr
computeFlowrate = CustomOperator(func=computeFlowrate,
                                 invars=(t, ),
                                 outvars=(flowrate, ))
correctFlowrate = FlowRateCorrection(
    implementation=Implementation.PYTHON,
    name="flowrate_correction",
    velocity=velo, vorticity=vorti,
    flowrate=flowrate,
    dt=dt,
    variables={velo: topo_nogh, vorti: topo_nogh},
    **extra_op_kwds)

#> outputs
penal.dump_inputs(fields=(sphere, ), frequency=outfrea, io_params=args.io_params)
correctFlowrate.dump_inputs(fields=(vorti, ), frequency=outfreq, io_params=args.io_params)
correctFlowrate.dump_outputs(fields=(velo,),  frequency=outfreq, io_params=args.io_params)

#> Operator to compute the infinite norm of the velocity
min_max_U = MinMaxFieldStatistics(name='min_max_U', field=velo,
        Finf=True, implementation=Implementation.PYTHON, variables={velo:npts},
        **extra_op_kwds)
#> Operator to compute the infinite norm of the vorticity
min_max_W = MinMaxFieldStatistics(name='min_max_W', field=vorti,
        Finf=True, implementation=Implementation.PYTHON, variables={vorti:npts},
        **extra_op_kwds)
#> Operator to compute the enstrophy
enstrophy_op = Enstrophy(
    name='enstrophy',
    vorticity=vorti, enstrophy=enstrophy, WdotW=wdotw,
    variables={vorti:topo_nogh, wdotw: topo_nogh},
    implementation=Implementation.PYTHON, **extra_op_kwds)

### Adaptive timestep operator
#TODO:move advection to GRAD_U
#TODO:add stretching criteria, based on a gradient
adapt_dt = AdaptiveTimeStep(dt, equivalent_CFL=True,
                            start_time=10*dt0)  # start adapting timestep at t=10*dt0
dt_cfl = adapt_dt.push_cfl_criteria(cfl=cfl,
                                    Finf=min_max_U.Finf,
                                    equivalent_CFL=True)
dt_advec = adapt_dt.push_advection_criteria(lcfl=lcfl,
                                            Finf=min_max_W.Finf,
                                            criteria=AdvectionCriteria.W_INF)
                                            #criteria=AdvectionCriteria.GRAD_U)
# dt_advec = adapt_dt.push_stretching_criteria(lcfl=lcfl,
#                                              gradFinf=grad_W.Finf,
#                                              criteria=StretchingCriteria.GRAD_U)

## Create the problem we want to solve and insert our
# directional splitting subgraph and the standard operators.
# The method dictionnary passed to this graph will be dispatched
# accross all operators contained in the graph.
method.update({SpaceDiscretization:   4,
               TimeIntegrator:        RK2,
               Remesh:                Remesh.L4_2,
               Interpolation:         Interpolation.LINEAR})
problem = Problem(method=method)
problem.insert(
    computeFlowrate,
    penal,
    splitting,
    diffuse,
    advec,
    absorption,
    poisson,
    correctFlowrate,
    enstrophy_op, min_max_U, min_max_W,
    adapt_dt
)
problem.build()

# # Create a simulation
# # (do not forget to specify the t and dt parameters here)
simu = Simulation(start=0., end=10.0,
                  dt0=dt0, t=t, dt=dt)
simu.write_parameters(t, dt_cfl, dt_advec, dt, enstrophy, flowrate,
                      min_max_U.Finf, min_max_W.Finf, adapt_dt.equivalent_CFL,
                      filename='parameters.txt', precision=8)

problem.initialize_field(vorti, formula=computeVort)
problem.initialize_field(velo, formula=computeVel)
problem.initialize_field(sphere, formula=computeSphere)

# Finally solve the problem
problem.solve(simu)

# Finalize
problem.finalize()
