"""Everything concerning GPU in hysop.

OpenCL sources are located in the cl_src directory and organized as follows
  - kernels/
  Contains kernels src
    - advection.cl
    - remeshing.cl
    - advection_and_remeshing.cl
    - rendering.cl
  Functions used by kernels
  - advection/
    - builtin.cl
    - builtin_noVec.cl
  - remeshing/
    - basic.cl
    - basic_noVec.cl
    - private.cl
    - weights.cl
    - weights_builtin.cl
    - weights_noVec.cl
    - weights_noVec_builtin.cl
  - common.cl

Some sources are parsed at build to handle several OpenCL features.
Other sources are generated and optimized at runtime.
see hysop.gpu.tools.parse_file
see hysop.backend.device.codegen

"""
import os
import pyopencl
import pyopencl.tools
import pyopencl.array
import pyopencl.characterize
import pyopencl.reduction
import pyopencl.clrandom
import pyopencl.elementwise
import pyopencl.scan

from hysop import __DEFAULT_PLATFORM_ID__, __DEFAULT_DEVICE_ID__, __VERBOSE__

## open cl underlying implementation
cl = pyopencl
"""open cl underlying implementation"""

clTools = pyopencl.tools
"""PyOpencl tools"""

clArray = pyopencl.array
"""PyOpenCL arrays"""

clRandom = pyopencl.clrandom
"""PyOpenCL random"""

clReduction = pyopencl.reduction
"""PyOpenCL reductions"""

clScan = pyopencl.scan
"""PyOpenCL scan"""

clElementwise = pyopencl.elementwise
"""PyOpenCL reductions"""

clCharacterize = pyopencl.characterize
"""PyOpenCL characterize"""

GPU_SRC = os.path.join(__path__[0], "cl_src", '')
"""Default path to cl kernels source files"""

KERNEL_DUMP_FOLDER='generated_kernels'
"""Default folder to dump debug opencl kernel sources"""

CL_PROFILE = False
"""Boolean, true to enable OpenCL profiling events to time computations"""

if __DEFAULT_PLATFORM_ID__ < len(cl.get_platforms()):
    default_cl_platform = cl.get_platforms()[__DEFAULT_PLATFORM_ID__]
else:
    default_cl_platform = cl.get_platforms()[0]
    msg= 'Warning: default opencl platform not available,'
    msg+=' switched to platform {}.'.format(default_cl_platform.name)
    print msg

if __DEFAULT_DEVICE_ID__ < len(default_cl_platform.get_devices()):
    default_cl_device   = default_cl_platform.get_devices()[__DEFAULT_DEVICE_ID__]
else:
    default_cl_device   = default_cl_platform.get_devices()[0]
    msg= 'Warning: default device not available,'
    msg+=' switched to device {}.'.format(default_cl_device.name)
    print msg

default_cl_context  = cl.Context(devices=[default_cl_device])
default_cl_queue    = cl.CommandQueue(context=default_cl_context, device=default_cl_device)

default_cl_allocator = clTools.ImmediateAllocator(queue=default_cl_queue, 
                                                    mem_flags=cl.mem_flags.READ_WRITE)
default_cl_pool = clTools.MemoryPool(default_cl_allocator)
