"""
"""
from hysop.constants import debug, DirectionLabels
from hysop import __VERBOSE__, __DEBUG__
from hysop.backend.device.opencl import cl, CL_PROFILE
from hysop.tools.profiler import FProfiler


class KernelListLauncher(object):
    """OpenCL kernel list launcher.

    Manage launching of OpenCL kernels as a list.
    """
    @debug
    def __init__(self, kernel, queue, gsize, lsize=None):
        """Create a kernel list launcher.

        Parameters
        ----------

        kernel : list
        queue : OpenCL command queue
        gsize : int
            OpenCL global size index.
        lsize : int, optional
            OpenCL local size index.
        """
        ## OpenCL Kernel list
        self.kernel = kernel
        #print [k.function_name for k in self.kernel]
        ## OpenCL command queue
        self.queue = queue
        ## OpenCL global size index.
        self.global_size = gsize
        ## OpenCL local size index.
        self.local_size = lsize
        if CL_PROFILE:
            if len(self.kernel) == 1:
                try:
                    self.profile = [FProfiler("OpenCL_" + k.function_name)
                                    for k in self.kernel]
                except AttributeError:
                    self.profile = [FProfiler("OpenCL_" + k.__name__)
                                    for k in self.kernel]
            else:
                self.profile = [
                    FProfiler("OpenCL_" + k.function_name + DirectionLabels[d])
                    for d, k in enumerate(self.kernel)]
        else:
            self.profile = []

    @debug
    def __call__(self, d, *args, **kwargs):
        """
        Launch a kernel.

        OpenCL global size and local sizes are not given in
        args. Class member are used.

        @param d : kernel index in kernel list.
        @param args : kernel arguments.
        @return OpenCL Event
        """
        return KernelListLauncher.launch_sizes_in_args(
            self, d, self.global_size[d], self.local_size[d], *args, **kwargs)

    def _kernel_name(self,d):
        try:
            kname = self.kernel[d].function_name
        except AttributeError:
            kname = self.kernel[d].__name__
        return kname


    @debug
    def launch_sizes_in_args(self, d, *args, **kwargs):
        """
        Launch a kernel.

        Opencl global and local sizes are given in args.

        @param d : kernel index in kernel list.
        @param args : kernel arguments.
        @return OpenCL Event.
        """
        if __VERBOSE__:
            try:
                print '  ',self.kernel[d].function_name, d, args[0], args[1]
                # print "\targs: ",args[2:]
                # print "\tkwargs: ", kwargs
            except AttributeError:
                print '  ', self.kernel[d].__name__
        evt = self.kernel[d](self.queue, *args, **kwargs)
        if CL_PROFILE:
            evt.wait()
            self.profile[d] += (evt.profile.end - evt.profile.start) * 1e-9
        if __VERBOSE__:
            try:
                evt.wait()
                status = evt.command_execution_status 
                if status < 0:
                    msg='\n/!\ Execution error {} while waiting kernel {} to finish /!\\\n'.format(status,self._kernel_name(d))
                    raise RuntimeError(msg)
            except Exception, e:
                msg='\n/!\ Unknown execution error while waiting kernel {} to finish /!\\\n'.format(self._kernel_name(d))
                print msg
                raise e


        return evt

    def function_name(self, d=None):
        """Prints OpenCL Kernels function names informations"""
        if d is not None:
            kernel = self.kernel[d]
            try:
                name = kernel.function_name
            except:
                name = kernel.__name__
            return name
        else:
            return [self.function_name(d)
                    for d in xrange(len(self.kernel))]


class OpenClKernelLauncher(KernelListLauncher):
    """
    OpenCL kernel launcher.

    Manage launching of one OpenCL kernel as a KernelListLauncher
    with a list of one kernel.
    """
    @debug
    def __init__(self, kernel, queue, gsize=None, lsize=None):
        """
        Create a OpenClKernelLauncher.

        Create a KernelListLauncher with a list of one kernel.

        @param kernel : kernel.
        @param queue : OpenCL command queue.
        @param gsize : OpenCL global size index.
        @param lsize : OpenCL local size index.
        """
        KernelListLauncher.__init__(self, [kernel], queue, [gsize], [lsize])

    @debug
    def launch_sizes_in_args(self, *args, **kwargs):
        """
        Launch the kernel.

        Opencl global and local sizes are given in args.

        @param args : kernel arguments.
        @return OpenCL Event.
        """
        return KernelListLauncher.launch_sizes_in_args(
            self, 0, *args, **kwargs)

    @debug
    def __call__(self, *args, **kwargs):
        """
        Launch the kernel.

        OpenCL global size and local sizes are not given in args.
        Class member are used.

        @param args : kernel arguments.
        @return OpenCL Event
        """
        return KernelListLauncher.__call__(self, 0, *args, **kwargs)

    def function_name(self):
        """Prints OpenCL Kernel function name informations"""
        res = KernelListLauncher.function_name(self, 0)
        return res
