"""
@file gpu_stretching.py

Stretching on GPU
"""
from hysop import __VERBOSE__, __DEBUG__
from hysop.constants import debug, np, DirectionLabels, hysop.core.mpi_REAL, ORDERMPI, \
    HYSOP_REAL, ORDER
from hysop.tools.numpywrappers import npw
from hysop.tools.io_utils import IO
from hysop.operator.discrete.discrete import DiscreteOperator, get_extra_args_from_method

from hysop.core.mpi import Wtime
from hysop.backend.device.opencl import cl
from hysop.backend.device.opencl.gpu_operator import GPUOperator
from hysop.backend.device.opencl.opencl_kernel   import OpenClKernelLauncher
from hysop.backend.device.opencl.opencl_discrete import OpenClDiscreteField
from hysop.tools.profiler   import FProfiler

from hysop.methods                 import TimeIntegrator, SpaceDiscretization, Formulation, Support
from hysop.numerics.odesolvers          import Euler, RK2, RK3, RK4
from hysop.numerics.finite_differences  import FDC2, FDC4
from hysop.operator.discrete.stretching import Conservative, GradUW

from hysop.backend.device.codegen.kernels.stretching import CachedStretchingKernel
from hysop.backend.device.opencl.opencl_tools import KernelAutotuner

class GPUStretching(DiscreteOperator, GPUOperator):

    _supported_time_integrators        = [Euler]
    _supported_stretching_formulations = [GradUW]
    _supported_space_discretizations   = [FDC2, FDC4]

    @debug
    def __init__(self, velocity, vorticity, **kwds):
        if (velocity  is None): raise ValueError('Set velocity  is None!')
        if (vorticity is None): raise ValueError('Set vorticity is None!')
        variables = [velocity, vorticity]

        if 'method' not in kwds:
            import hysop.default_methods as default
            kwds['method'] = default.GPU_STRETCHING
        self._check_method(kwds['method'])

        super(GPUStretching, self).__init__(variables=variables, **kwds)
        self.input  = [velocity, vorticity]
        self.output = [vorticity]
        
        GPUOperator.__init__(self,
            platform_id = get_extra_args_from_method(self, 'platform_id', None),
            device_id   = get_extra_args_from_method(self, 'device_id',   None),
            device_type = get_extra_args_from_method(self, 'device_type', None),
            **kwds)
        
        # Build options
        build_opts = []
        build_opts += get_extra_args_from_method(self, 'build_opts', [])
        self.build_opts = filter(None, set(build_opts))
        
        # Kernel autotuner and caching options
        self.autotuner_runs    = get_extra_args_from_method(self, 'autotuner_runs', 10)
        self.force_renew_cache = get_extra_args_from_method(self, 'force_renew_cache', False)

        # discrete fields
        self.vorticity = vorticity
        self.velocity  = velocity
        
        # order of spatial scheme
        self.order = 2 if self.method[SpaceDiscretization] is FDC2 else 4
        
        # Worksize handling
        self._cl_work_size = 0

        ## GPU allocations
        for field in variables:
            OpenClDiscreteField.fromField(self.cl_env, field,
                                    self.gpu_precision, simple_layout=False)
            if field.allocate():
                self.size_global_alloc += field.mem_size

        topo = self.velocity.topology
        self._cutdir_list = np.where(topo.cutdir)[0].tolist()
        self._comm        = topo.comm
        self._comm_size   = self._comm.Get_size()
        self._comm_rank   = self._comm.Get_rank()
        if self._comm_size > 1:
            self._compute = None
            raise RuntimeError('GPU Stretching does not support mutiple processes for now!')
        else:
            self._compute = self._compute_stretching
        
        # generate source and compile kernels
        self._gen_cl_src()

    def _check_method(self, method):
        if Support not in method.keys() or method[Support].lower().find('gpu')<0:
            msg = 'GPU Stretching operator must be called with explicit GPU Support!'
            raise ValueError(msg)

        if TimeIntegrator not in method.keys():
            msg = 'No time integrator specified for the GPU Stretching Operator!'
            raise ValueError(msg)
        elif method[TimeIntegrator] not in self._supported_time_integrators:
            msg =  'Asked for a \'{}\' time integrator '.format(method[TimeIntegrator])
            msg += 'but only {} time integrators have been implemented in the GPU Stretching Operator!'.format(self._supported_time_integrators)
            raise NotImplementedError(msg)
        
        if SpaceDiscretization not in method.keys():
            msg = 'No space discretization specified for the GPU Stretching Operator!'
            raise ValueError(msg)
        elif method[SpaceDiscretization] not in self._supported_space_discretizations:
            msg =  'Asked for a \'{}\' space discretization '.format(method[SpaceDiscretization])
            msg += 'but only {} space discretizations have been implemented in the GPU Stretching Operator!'.format(self._supported_space_discretizations)
            raise NotImplementedError(msg)

        if Formulation not in method.keys():
            msg = 'No formulation specified for the GPU Stretching Operator!'
            raise ValueError(msg)
        elif method[Formulation] not in self._supported_stretching_formulations:
            msg =  'Asked for a \'{}\' stretching formulation '.format(method[Formulation])
            msg += 'but only {} formulations have been implemented in the GPU Stretching Operator!'.format(self._supported_stretching_formulations)
            raise NotImplementedError(msg)

    def _gen_cl_src(self):
        cl_env   = self.cl_env
        context  = cl_env.ctx
        device   = cl_env.device
        platform = cl_env.platform
        typegen  = cl_env.typegen

        topo     = self.velocity.topology
        work_dim = self.dim
        mesh     = topo.mesh
        gwi      = mesh.local_resolution
        
        build_opts        = self.build_opts
        autotuner_runs    = self.autotuner_runs
        force_renew_cache = self.force_renew_cache
        
        dt = typegen.make_floatn([0.01],1)
        
        from hysop.backend.device.codegen.structs.mesh_info import MeshInfoStruct
        mesh_info = MeshInfoStruct.build_instance_from_mesh(typegen, mesh)
        mesh_info_buffer = cl.Buffer(context, cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
                hostbuf=mesh_info)

        
        kernel_args    = [dt]                                               \
                        + self.vorticity.gpu_data                           \
                        + self.velocity.gpu_data                            \
                        + [mesh_info_buffer]                                \
                        + [None] #local memory buffer
        
        name = CachedStretchingKernel.codegen_name(typegen.fbtype, work_dim)
        autotuner = KernelAutotuner(name,work_dim,typegen.fbtype,build_opts,
                nruns=autotuner_runs,force_renew_cache=force_renew_cache)
        autotuner.add_filter('3d_shape', autotuner.min_workitems_per_direction)

        (gwi, lwi, stats) = autotuner.bench(context, platform, device, gwi, kernel_args, 
                kernel_generator=self._gen_and_build_kernel, dump_src=False, verbose=False,
                min_local_size=4, min_load_factor=4)

        (kernel, kernel_args, cached_bytes) = self._gen_and_build_kernel(lwi, gwi, kernel_args, 
                dump_src=True, verbose=True)


        self.kernel_args = kernel_args
        self.size_local_alloc += cached_bytes
        
        kernels = {}
        kernels['stretching'] = OpenClKernelLauncher(kernel, cl_env.queue, gwi, lwi)
        self.kernels = kernels

    def _gen_and_build_kernel(self, local_size, global_size, kernel_args, 
            dump_src=False, build_opts=None, verbose=False, **kargs):
        cl_env  = self.cl_env
        context = cl_env.ctx
        device  = cl_env.device
        typegen = cl_env.typegen

        work_dim      = self.dim 
        symbolic_mode = __VERBOSE__ or __DEBUG__
        order         = self.order
        known_vars    = { 'local_size':  local_size }

        codegen = CachedStretchingKernel(typegen=typegen,
                work_dim=work_dim,
                order=order,
                device=device, 
                context=context,
                known_vars=known_vars,
                symbolic_mode=symbolic_mode)
        
        src = codegen.__str__()
        if dump_src:
            dump_folder=IO.default_path()+'/generated_kernels'
            codegen.to_file(dump_folder, codegen.name+'.cl')
            if __VERBOSE__ or __DEBUG__:
                print '{} source dumped to {}/{}.'.format(codegen.name, dump_folder, codegen.name+'.cl')
        
        cached_bytes = codegen.cache_alloc_bytes(local_size)
        local_mem       = cl.LocalMemory(cached_bytes)
        kernel_args[-1] = local_mem

        build_opts = [] if (build_opts is None) else build_opts

        prg        = cl_env.build_raw_src(src, build_opts, force_verbose=verbose)
        kernel     = prg.all_kernels()[0]

        return (kernel, kernel_args, cached_bytes)

    def _compute_stretching(self, simulation):
        dt = self.cl_env.typegen.make_floatn(simulation.time_step,1)
        
        kernel = self.kernels['stretching']
        kernel_args = self.kernel_args
        kernel_args[0] = dt

        input_events   = [evt for input in self.input for evt in input.events]
        stretching_evt = kernel(*kernel_args, wait_for=input_events)
        output_events  = [stretching_evt]
        
        for var in self.variables:
            var.events += output_events

    def apply(self, simulation):
        self._compute(simulation)

    def get_profiling_info(self):
        for k in self.kernels.values():
            if k is not None:
                for p in k.profile:
                    self.profiler += p
