"""Testing advection kernels with null velocity.
"""
from hysop.domain.box import Box
from hysop.fields.continuous_field import Field
from hysop.operator.advection import Advection
from hysop.constants import np, HYSOP_REAL
from hysop.problem.simulation import Simulation
from hysop.methods import TimeIntegrator, Interpolation, Remesh, \
    Support, Splitting, Precision
from hysop.numerics.odesolvers import RK2
from hysop.numerics.interpolation.interpolation import Linear
from hysop.numerics.remeshing import L2_1, L4_2, L6_3, M8Prime
from hysop.tools.parameters import Discretization
from hysop.tools.numpywrappers import npw
from hysop.problem.simulation import O2, O2FULLHALF


d2d = Discretization([33, 33])
d3d = Discretization([17, 17, 17])
m1k = {TimeIntegrator: RK2, Interpolation: Linear,
       Remesh: L4_2, Support: 'gpu_1k', Splitting: O2,
       Precision: HYSOP_REAL}
m2k = {TimeIntegrator: RK2, Interpolation: Linear,
       Remesh: L4_2, Support: 'gpu_2k', Splitting: O2,
       Precision: HYSOP_REAL}


def run_advection(discr, vector_field, method=None):
    """Create advection operator, ref operator
    and fields, run scales and python advection,
    compare results.

    Parameters
    ----------
    dicr : :class:`~hysop.tools.parameters.Discretization`
        chosen discretization for operator --> set domain dimension
    vector_field: bool
        True to advect a vector field, else scalar field.
    method : dictionnary
        Set scales remeshing type.
    """
    dimension = len(discr.resolution)
    box = Box(length=[1., ] * dimension, origin=[0., ] * dimension)
    scal = Field(domain=box, name='Scalar', is_vector=vector_field)
    scal_ref = Field(domain=box, name='Scalar_ref', is_vector=vector_field)
    velo = Field(domain=box, name='Velocity',
                 formula=lambda x, y, z, t: (0., 0., 0.), is_vector=True,
                 vectorize_formula=True)
    # build gpu advection
    if method is None:
        method = m1k

    advec = Advection(velocity=velo,
                      advected_fields=scal,
                      discretization=discr,
                      method=method)
    advec.discretize()
    advec.setup()
    
    # Get and randomize discrete fields
    topo = advec.advected_fields_topology()
    scal_d = scal.randomize(topo)
    assert (velo.norm(topo) == 0).all()
    ic = topo.mesh.compute_index
    # copy data for reference
    scal_ref.copy(scal, topo)
    scal_ref_d = scal_ref.discretize(topo)
    topo_velo = advec.velocity_topology()
    velo_d = velo.discretize(topo_velo)
    # transfer data to gpu
    scal_d.toDevice()
    velo_d.toDevice()
    scal_d.wait()
    velo_d.wait()

    advec.apply(Simulation(start=0., end=0.1, nb_iter=1))
    # Get data back
    scal_d.toHost()
    scal_d.wait()
    advec.finalize()
    for d in xrange(len(scal_d.data)):
        assert np.allclose(scal_ref_d.data[d][ic], scal_d.data[d][ic])


def test_2D_m6_1k():
    """Test M6 remeshing formula in 2D, 1 kernel,
    o2 splitting.
    """
    run_advection(d2d, False, m1k)


# def test_2D_m6_2k():
#     """Test M6 remeshing formula in 2D, 2 kernels,
#     o2 splitting.
#     """
#     run_advection(d2d, False, m2k)


# def test_2D_m6_1k_sFH():
#     """Test M6 remeshing formula in 2D, 1 kernel,
#     o2 full-half splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: L4_2, Support: 'gpu_1k', Splitting: O2FULLHALF,
#             Precision: HYSOP_REAL}
#     run_advection(d2d, False, meth)


# def test_2D_m6_2k_sFH():
#     """Test M6 remeshing formula in 2D, 2 kernels,
#     o2 full-half splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: L4_2, Support: 'gpu_2k', Splitting: O2FULLHALF,
#             Precision: HYSOP_REAL}
#     run_advection(d2d, False, meth)


# def test_3D_m6_1k():
#     """Test M6 remeshing formula in 3D, 1 kernel,
#     o2 splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: L4_2, Support: 'gpu_1k', Splitting: O2,
#             Precision: HYSOP_REAL}
#     run_advection(d3d, False, meth)


# def test_3D_m6_2k():
#     """Test M6 remeshing formula in 3D, 2 kernels,
#     o2 splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: L4_2, Support: 'gpu_2k', Splitting: O2,
#             Precision: HYSOP_REAL}
#     run_advection(d3d, False, meth)


# def test_3D_m6_1k_sFH():
#     """Test M6 remeshing formula in 3D, 1 kernel,
#     o2 fh splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: L4_2, Support: 'gpu_1k', Splitting: O2FULLHALF,
#             Precision: HYSOP_REAL}
#     run_advection(d3d, False, meth)


# def test_3D_m6_2k_sFH():
#     """Test M6 remeshing formula in 3D, 2 kernels,
#     o2 fh splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: L4_2, Support: 'gpu_2k', Splitting: O2FULLHALF,
#             Precision: HYSOP_REAL}
#     run_advection(d3d, False, meth)


# def test_2D_m4_1k():
#     """Test M4 remeshing formula in 2D, 1 kernel,
#     o2 splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: L2_1, Support: 'gpu_1k', Splitting: O2,
#             Precision: HYSOP_REAL}
#     run_advection(d2d, False, meth)


# def test_2D_m4_2k():
#     """Test M4 remeshing formula in 2D, 2 kernels,
#     o2 splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: L2_1, Support: 'gpu_2k', Splitting: O2,
#             Precision: HYSOP_REAL}
#     run_advection(d2d, False, meth)


# def test_2D_m4_1k_sFH():
#     """Test M4 remeshing formula in 2D, 1 kernel,
#     o2 full-half splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: L2_1, Support: 'gpu_1k', Splitting: O2FULLHALF,
#             Precision: HYSOP_REAL}
#     run_advection(d2d, False, meth)


# def test_2D_m4_2k_sFH():
#     """Test M4 remeshing formula in 2D, 2 kernels,
#     o2 full-half splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: L2_1, Support: 'gpu_2k', Splitting: O2FULLHALF,
#             Precision: HYSOP_REAL}
#     run_advection(d2d, False, meth)


# def test_3D_m4_1k():
#     """Test M4 remeshing formula in 3D, 1 kernel,
#     o2 splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: L2_1, Support: 'gpu_1k', Splitting: O2,
#             Precision: HYSOP_REAL}
#     run_advection(d3d, False, meth)


# def test_3D_m4_2k():
#     """Test M4 remeshing formula in 3D, 2 kernels,
#     o2 splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: L2_1, Support: 'gpu_2k', Splitting: O2,
#             Precision: HYSOP_REAL}
#     run_advection(d3d, False, meth)


# def test_3D_m4_1k_sFH():
#     """Test M4 remeshing formula in 3D, 1 kernel,
#     o2 fh splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: L2_1, Support: 'gpu_1k', Splitting: O2FULLHALF,
#             Precision: HYSOP_REAL}
#     run_advection(d3d, False, meth)


# def test_3D_m4_2k_sFH():
#     """Test M4 remeshing formula in 3D, 2 kernels,
#     o2 fh splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: L2_1, Support: 'gpu_2k', Splitting: O2FULLHALF,
#             Precision: HYSOP_REAL}
#     run_advection(d3d, False, meth)


# def test_2D_m8_1k():
#     """Test M8 remeshing formula in 2D, 1 kernel,
#     o2 splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: M8Prime, Support: 'gpu_1k', Splitting: O2,
#             Precision: HYSOP_REAL}
#     run_advection(d2d, False, meth)


# def test_2D_m8_2k():
#     """Test M8 remeshing formula in 2D, 2 kernels,
#     o2 splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: M8Prime, Support: 'gpu_2k', Splitting: O2,
#             Precision: HYSOP_REAL}
#     run_advection(d2d, False, meth)


# def test_2D_m8_1k_sFH():
#     """Test M8 remeshing formula in 2D, 1 kernel,
#     o2 full-half splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: M8Prime, Support: 'gpu_1k', Splitting: O2FULLHALF,
#             Precision: HYSOP_REAL}
#     run_advection(d2d, False, meth)


# def test_2D_m8_2k_sFH():
#     """Test M8 remeshing formula in 2D, 2 kernels,
#     o2 full-half splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: M8Prime, Support: 'gpu_2k', Splitting: O2FULLHALF,
#             Precision: HYSOP_REAL}
#     run_advection(d2d, False, meth)


# def test_3D_m8_1k():
#     """Test M8 remeshing formula in 3D, 1 kernel,
#     o2 splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: M8Prime, Support: 'gpu_1k', Splitting: O2,
#             Precision: HYSOP_REAL}
#     run_advection(d3d, False, meth)


# def test_3D_m8_2k():
#     """Test M8 remeshing formula in 3D, 2 kernels,
#     o2 splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: M8Prime, Support: 'gpu_2k', Splitting: O2,
#             Precision: HYSOP_REAL}
#     run_advection(d3d, False, meth)


# def test_3D_m8_1k_sFH():
#     """Test M8 remeshing formula in 3D, 1 kernel,
#     o2 fh splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: M8Prime, Support: 'gpu_1k', Splitting: O2FULLHALF,
#             Precision: HYSOP_REAL}
#     run_advection(d3d, False, meth)


# def test_3D_m8_2k_sFH():
#     """Test M8 remeshing formula in 3D, 2 kernels,
#     o2 fh splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: M8Prime, Support: 'gpu_2k', Splitting: O2FULLHALF,
#             Precision: HYSOP_REAL}
#     run_advection(d3d, False, meth)


# def test_2D_l6_1k():
#     """Test L6 remeshing formula in 2D, 1 kernel,
#     o2 splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: L6_3, Support: 'gpu_1k', Splitting: O2,
#             Precision: HYSOP_REAL}
#     run_advection(d2d, False, meth)


# def test_2D_l6_2k():
#     """Test L6 remeshing formula in 2D, 2 kernels,
#     o2 splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: L6_3, Support: 'gpu_2k', Splitting: O2,
#             Precision: HYSOP_REAL}
#     run_advection(d2d, False, meth)


# def test_2D_l6_1k_sFH():
#     """Test L6 remeshing formula in 2D, 1 kernel,
#     o2 full-half splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: L6_3, Support: 'gpu_1k', Splitting: O2FULLHALF,
#             Precision: HYSOP_REAL}
#     run_advection(d2d, False, meth)


# def test_2D_l6_2k_sFH():
#     """Test L6 remeshing formula in 2D, 2 kernels,
#     o2 full-half splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: L6_3, Support: 'gpu_2k', Splitting: O2FULLHALF,
#             Precision: HYSOP_REAL}
#     run_advection(d2d, False, meth)


# def test_3D_l6_1k():
#     """Test L6 remeshing formula in 3D, 1 kernel,
#     o2 splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: L6_3, Support: 'gpu_1k', Splitting: O2,
#             Precision: HYSOP_REAL}
#     run_advection(d3d, False, meth)


# def test_3D_l6_2k():
#     """Test L6 remeshing formula in 3D, 2 kernels,
#     o2 splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: L6_3, Support: 'gpu_2k', Splitting: O2,
#             Precision: HYSOP_REAL}
#     run_advection(d3d, False, meth)


# def test_3D_l6_1k_sFH():
#     """Test L6 remeshing formula in 3D, 1 kernel,
#     o2 fh splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: L6_3, Support: 'gpu_1k', Splitting: O2FULLHALF,
#             Precision: HYSOP_REAL}
#     run_advection(d3d, False, meth)


# def test_3D_l6_2k_sFH():
#     """Test L6 remeshing formula in 3D, 2 kernels,
#     o2 fh splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: L6_3, Support: 'gpu_2k', Splitting: O2FULLHALF,
#             Precision: HYSOP_REAL}
#     run_advection(d3d, False, meth)


# def test_rectangular_domain2D():
#     """Test remeshing formula in 2D, with different resolutions in each dir,
#     o2 splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: L4_2, Support: 'gpu_1k', Splitting: O2,
#             Precision: HYSOP_REAL}
#     run_advection(Discretization([65, 33]), False, meth)


# def test_rectangular_domain3D():
#     """Test remeshing formula in 3D, with different resolutions in each dir,
#     o2 splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: L4_2, Support: 'gpu_1k', Splitting: O2,
#             Precision: HYSOP_REAL}
#     run_advection(Discretization([65, 33, 17]), False, meth)


# def test_vector_2D():
#     """Test remeshing formula in 2D, advection of a vector field,
#     o2 splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: L4_2, Support: 'gpu_1k', Splitting: O2,
#             Precision: HYSOP_REAL}
#     run_advection(Discretization([129, 129]), True, meth)


# def test_vector_3D():
#     """Test remeshing formula in 3D, advection of a vector field,
#     o2 splitting.
#     """
#     meth = {TimeIntegrator: RK2, Interpolation: Linear,
#             Remesh: L4_2, Support: 'gpu_1k', Splitting: O2,
#             Precision: HYSOP_REAL}
#     run_advection(d3d, False, meth)
