"""Finite difference schemes

.. currentmodule hysop.numerics.finite_differences

* :class:`~FDC2` : first derivative, 2nd order centered scheme
* :class:`~FDC4`: first derivative, 4th order centered scheme
* :class:`~FD2C2`: second derivative, 2nd order centered scheme
* :class `~FDC`: generic centered scheme
* :class:`~FiniteDifference`, abstract base class.


Usage:

.. code::

    # indices and space step from a predefined mesh
    indices = topo.mesh.compute_index
    step = topo.mesh.space_step
    # init scheme
    scheme = FDC4(step, indices)
    # result = some array of the same shape as field.
    # compute dfield/d direction in result
    result = scheme(field, direction, result)
    # or (with more assert and checks on inputs)
    scheme.compute(field, direction, result)
    # or result += dfield/d direction
    scheme.compute_and_add(field, direction, result)


Note
----
* fd are computed only on points described by indices :
  result[ind] = diff(field[ind]),
  which means that result and field must have the same shape.
* You can also use a 'reduced' output result to minimize memory.
  Use 'reduce_output_shape=True' in scheme init. In that case,
  result must be of a shape corresponding to indices 'shape'.

  Example:

.. code::

  scheme = FDC4(step, indices, reduce_output_shape=True)
  # if indices is slice(4,10) in each direction
  result = npw.zeros((6,6,6))
  ...

"""
from abc import ABCMeta, abstractmethod
from hysop.constants import debug
from hysop.numerics.stencil import Stencil, StencilGenerator
from hysop.numerics.stencil import CenteredStencil, CenteredStencilGenerator
from hysop.tools.numpywrappers import npw
import numpy as np


class FiniteDifference(object):
    """Describe and apply a finite difference scheme to compute
    1st or second derivative of a variable saved in a numpy array.

    Usage :

    >> step = topo.mesh_space_step
    >> scheme = FDC4(step, topo.mesh.compute_index)

    For a given numpy array (obviously discretized on the topo
    used to compute indices), to compute result as the derivative
    of tab according to dir:

    >> scheme.compute(tab, dir, result)

    or result += the derivative of tab according to dir:

    >> scheme.compute_and_add(tab, dir, result)
    
    >>> from hysop import Box
    >>> domain = Box()

    Notes FP :
    Compute method is much more performant than compute and add
    and needs less memory. But in some cases (when fd results need
    to be accumulate into a field) compute_and_add may be useful
    to limit memory usage, if required.
    See Global_tests/testPerfAndMemForFD_and_div.py for perf results.
    """

    __metaclass__ = ABCMeta

    @debug
    def __new__(cls, *args, **kw):
        return object.__new__(cls, *args, **kw)

    def __init__(self, step, indices, output_indices=None):
        """

        Parameters
        ----------
        step : list or array of int
            resolution of the mesh
        indices : list of slices
            Represents the local mesh on which finite-differences
            will be applied, like compute_index
            in :class:`~hysop.domain.mesh.Mesh`.
        output_indices : list of slices, optional
            set which indices of result will be modified. By default,
            output_indices = indices

        Attributes
        ----------
        ghosts_layer_size : int, static
            minimal number of points required inside the ghost layer
            for this scheme.

        """

        step = np.asarray(step)
        #  dim of the field on which scheme will be applied
        # (i.e dim of the domain)
        self._dim = step.size
        self._m1 = None
        self._a1 = None
        self._m2 = None
        self._a2 = None
        self._coeff = None
        # List of slices representing the mesh on which fd scheme is applied
        self.indices = indices
        if output_indices is None or output_indices is False:
            # This is the default case. It means that
            # result shape is unknown. It will
            # be the same as input tab shape during call
            # and input/output indices are the same:
            # we compute result[indices] = scheme(tab[indices])
            self.output_indices = self.indices
        elif isinstance(output_indices, list):
            # In that case, result shape remains unknown but
            # the position where result is written is not the
            # same as input indices.
            # We compute:
            # result[output_indices] = scheme(tab[indices])
            # Warning : result shape/output_indices compliance
            # is not checked!
            self.output_indices = output_indices

        elif output_indices is True:
            # Here, output shape is fixed to the minimal required
            # shape, computed with indices.
            # So, we compute:
            # result[...] = scheme(tab[indices])
            result_shape = tuple([indices[i].stop - indices[i].start
                                  for i in xrange(len(indices))])
            self.output_indices = [slice(0, result_shape[i])
                                   for i in xrange(len(indices))]
        else:
            raise ValueError('Invalid value for indices_out!')

        self._step = step
        self._compute_indices(step)

    @abstractmethod
    def _compute_indices(self, step):
        """Internal index lists and fd coeff computation

        """

    @abstractmethod
    def __call__(self, tab, cdir, result):
        """Apply FD scheme.
        """

    def compute(self, tab, cdir, result):
        """Apply FD scheme. Result is overwritten.

        Parameters
        ----------
        tab : numpy array
            input field
        cdir : int
            direction of differentiation
        result : numpy array
            in/out, derivative of tab
        """
        assert result is not tab
        assert result.__class__ is np.ndarray
        assert tab.__class__ is np.ndarray
        result = self.__call__(tab, cdir, result)
        return result

    @abstractmethod
    def compute_and_add(self, tab, cdir, result, work):
        """Apply FD scheme and add the result inplace.

        Parameters
        ----------
        tab : numpy array
            input field
        cdir : int
            direction of differentiation
        result : numpy array
            in/out, abs(derivative of tab) + result
        work : numpy array
            internal workspace. Must be 1D and of the same
            size as result.
        """

    @abstractmethod
    def compute_and_add_abs(self, tab, cdir, result, work):
        """Apply FD scheme and add np.abs of the derivative inplace.

        Parameters
        ----------
        tab : numpy array
            input field
        cdir : int
            direction of differentiation
        result : numpy array
            in/out, derivative of tab + result
        work : numpy array
            internal workspace. Must be 1D and of the same
            size as result.
        """

class FDkCp(FiniteDifference):
    """
    k-th derivative, centered scheme, p-th order.

    """
    
    def __init__(self,k,p,*args,**kargs):
        super(FDkCp,self).__init__(*args,**kargs)
        
        csg     = CenteredStencilGenerator()
        stencil = csg.generate_exact_stencil(order=p,derivative=k)
        
        lghosts = stencil.L
        rghosts = stencil.R
        
        self.stencil = stencil
        self.right_ghost_layer_size = rghosts
        self.left_ghosts_layer_size = lghosts
        self.ghosts_layer_size      = np.maximum(lghosts, rghosts)
    
    def __call__(self, tab, cdir, result):
        stencil = self.stencil
        svars = {stencil.dx:self._step[cdir]}
         
        self.output_indices = [idx for idx in self.output_indices]
        result[self.output_indices] = 0
        for (offset,coeff) in self.stencil.iteritems(svars):
            result[self.output_indices] += coeff*tab[self._indices(offset)]
        return result
    
    def _compute_indices(self, step):
        pass
    
    def compute_and_add(self):
        pass
    def compute_and_add_abs(self):
        pass



class FDC2(FiniteDifference):
    """
    1st derivative, centered scheme, 2nd order.

    """

    ghosts_layer_size = 1

    def _compute_indices(self, step):

        self._coeff = npw.asarray(1. / (2. * step))
        self._m1 = []
        self._a1 = []
        for dim in xrange(self._dim):
            self._m1.append(list(self.indices))
            self._m1[dim][dim] = slice(self.indices[dim].start - 1,
                                       self.indices[dim].stop - 1,
                                       self.indices[dim].step)
            self._a1.append(list(self.indices))
            self._a1[dim][dim] = slice(self.indices[dim].start + 1,
                                       self.indices[dim].stop + 1,
                                       self.indices[dim].step)

    def __call__(self, tab, cdir, result):
        result[self.output_indices] = tab[self._a1[cdir]]
        result[self.output_indices] -= tab[self._m1[cdir]]
        result[self.output_indices] *= self._coeff[cdir]
        return result

    def compute_and_add(self, tab, cdir, result, work):
        """Apply FD scheme and add the derivative inplace.

        Parameters
        ----------
        tab : numpy array
            input field
        cdir : int
            direction of differentiation
        result : numpy array
            in/out, derivative of tab + result
        work : numpy array
            internal workspace. Must be 1D and of the same
            size as result.

        """
        assert result.__class__ is np.ndarray
        assert tab.__class__ is np.ndarray
        assert work.size == result.size
        wk = work.reshape(result.shape)
        np.subtract(tab[self._a1[cdir]], tab[self._m1[cdir]],
                    wk[self.output_indices])
        np.multiply(wk[self.output_indices], self._coeff[cdir],
                    wk[self.output_indices])
        np.add(wk[self.output_indices], result[self.output_indices],
               result[self.output_indices])
        return result

    def compute_and_add_abs(self, tab, cdir, result, work):
        """Apply FD scheme and add np.abs of the derivative inplace.

        Parameters
        ----------
        tab : numpy array
            input field
        cdir : int
            direction of differentiation
        result : numpy array
            in/out, derivative of tab + result
        work : numpy array
            internal workspace. Must be 1D and of the same
            size as result.

        """
        assert result.__class__ is np.ndarray
        assert tab.__class__ is np.ndarray
        assert work.__class__ is np.ndarray
        assert work is not result
        wk = work.reshape(result.shape)
        np.subtract(tab[self._a1[cdir]], tab[self._m1[cdir]],
                    wk[self.output_indices])
        np.multiply(wk[self.output_indices], self._coeff[cdir],
                    wk[self.output_indices])
        np.abs(wk[self.output_indices], wk[self.output_indices])
        np.add(wk[self.output_indices], result[self.output_indices],
               result[self.output_indices])
        return result


class FD2C2(FiniteDifference):
    """
    Second derivative, centered scheme, 2nd order.
    """

    ghosts_layer_size = 1

    def _compute_indices(self, step):

        self._m1 = []
        self._a1 = []
        self._coeff = npw.asarray(1. / (step * step))
        for dim in xrange(self._dim):
            self._m1.append(list(self.indices))
            self._m1[dim][dim] = slice(self.indices[dim].start - 1,
                                       self.indices[dim].stop - 1,
                                       self.indices[dim].step)
            self._a1.append(list(self.indices))
            self._a1[dim][dim] = slice(self.indices[dim].start + 1,
                                       self.indices[dim].stop + 1,
                                       self.indices[dim].step)

    def __call__(self, tab, cdir, result):
        result[self.output_indices] = tab[self.indices]
        result[self.output_indices] *= -2
        result[self.output_indices] += tab[self._a1[cdir]]
        result[self.output_indices] += tab[self._m1[cdir]]
        result[self.output_indices] *= self._coeff[cdir]
        return result

    def compute_and_add(self, tab, cdir, result, work):
        """Apply FD scheme and add the result inplace.

        Parameters
        ----------
        tab : numpy array
            input field
        cdir : int
            direction of differentiation
        result : numpy array
            in/out, derivative of tab + result
        work : numpy array
            internal workspace. Must be 1D and of the same
            size as result.

        """
        assert result.__class__ is np.ndarray
        assert tab.__class__ is np.ndarray
        wk = work.reshape(result.shape)
        np.multiply(tab[self.indices], -2., wk[self.output_indices])
        np.add(tab[self._a1[cdir]], wk[self.output_indices],
               wk[self.output_indices])
        np.add(tab[self._m1[cdir]], wk[self.output_indices],
               wk[self.output_indices])
        np.multiply(wk[self.output_indices], self._coeff[cdir],
                    wk[self.output_indices])
        np.add(wk[self.output_indices], result[self.output_indices],
               result[self.output_indices])
        return result

    def compute_and_add_abs(self, tab, cdir, result, work):
        """Apply FD scheme and add np.abs of the derivative inplace.

        Parameters
        ----------
        tab : numpy array
            input field
        cdir : int
            direction of differentiation
        result : numpy array
            in/out, derivative of tab + result
        work : numpy array
            internal workspace. Must be 1D and of the same
            size as result.

        """
        assert result.__class__ is np.ndarray
        assert tab.__class__ is np.ndarray
        wk = work.reshape(result.shape)
        np.multiply(tab[self.indices], -2., wk[self.output_indices])
        np.add(tab[self._a1[cdir]], wk[self.output_indices],
               wk[self.output_indices])
        np.add(tab[self._m1[cdir]], wk[self.output_indices],
               wk[self.output_indices])
        np.multiply(wk[self.output_indices], self._coeff[cdir],
                    wk[self.output_indices])
        np.abs(wk[self.output_indices], wk[self.output_indices])
        np.add(wk[self.output_indices], result[self.output_indices],
               result[self.output_indices])

        return result


class FDC4(FiniteDifference):
    """
    1st derivative, centered scheme, 4th order.
    """

    ghosts_layer_size = 2

    def _compute_indices(self, step):
        self._m1 = []
        self._m2 = []
        self._a1 = []
        self._a2 = []
        # FD scheme coefficients
        self._coeff = npw.asarray(1. / (12. * step))
        for dim in xrange(self._dim):
            self._m1.append(list(self.indices))
            self._m1[dim][dim] = slice(self.indices[dim].start - 1,
                                       self.indices[dim].stop - 1,
                                       self.indices[dim].step)
            self._m2.append(list(self.indices))
            self._m2[dim][dim] = slice(self.indices[dim].start - 2,
                                       self.indices[dim].stop - 2,
                                       self.indices[dim].step)
            self._a1.append(list(self.indices))
            self._a1[dim][dim] = slice(self.indices[dim].start + 1,
                                       self.indices[dim].stop + 1,
                                       self.indices[dim].step)
            self._a2.append(list(self.indices))
            self._a2[dim][dim] = slice(self.indices[dim].start + 2,
                                       self.indices[dim].stop + 2,
                                       self.indices[dim].step)

    def __call__(self, tab, cdir, result):
        result[self.output_indices] = tab[self._a1[cdir]]
        result[self.output_indices] -= tab[self._m1[cdir]]
        result[self.output_indices] *= 8
        result[self.output_indices] += tab[self._m2[cdir]]
        result[self.output_indices] -= tab[self._a2[cdir]]
        result[self.output_indices] *= self._coeff[cdir]
        return result

    def compute_and_add(self, tab, cdir, result, work):
        """Apply FD scheme and add the result inplace.

        Parameters
        ----------
        tab : numpy array
            input field
        cdir : int
            direction of differentiation
        result : numpy array
            in/out, derivative of tab + result
        work : numpy array
            internal workspace. Must be 1D and of the same
            size as result.

        """
        assert result.__class__ is np.ndarray
        assert tab.__class__ is np.ndarray
        assert work is not result
        assert work.size == result.size
        wk = work.reshape(result.shape)
        np.subtract(tab[self._a1[cdir]], tab[self._m1[cdir]],
                    wk[self.output_indices])
        np.multiply(wk[self.output_indices], 8.,
                    wk[self.output_indices])
        np.add(wk[self.output_indices], tab[self._m2[cdir]],
               wk[self.output_indices])
        np.subtract(wk[self.output_indices], tab[self._a2[cdir]],
                    wk[self.output_indices])
        np.multiply(wk[self.output_indices], self._coeff[cdir],
                    wk[self.output_indices])
        np.add(wk[self.output_indices], result[self.output_indices],
               result[self.output_indices])
        return result

    def compute_and_add_abs(self, tab, cdir, result, work):
        """Apply FD scheme and add np.abs of the derivative inplace.

        Parameters
        ----------
        tab : numpy array
            input field
        cdir : int
            direction of differentiation
        result : numpy array
            in/out, derivative of tab + result
        work : numpy array
            internal workspace. Must be 1D and of the same
            size as result.

        """
        assert result.__class__ is np.ndarray
        assert tab.__class__ is np.ndarray
        assert work is not result
        assert work.size == result.size
        wk = work.reshape(result.shape)
        np.subtract(tab[self._a1[cdir]], tab[self._m1[cdir]],
                    wk[self.output_indices])
        np.multiply(wk[self.output_indices], 8.,
                    wk[self.output_indices])
        np.add(wk[self.output_indices], tab[self._m2[cdir]],
               wk[self.output_indices])
        np.subtract(wk[self.output_indices], tab[self._a2[cdir]],
                    wk[self.output_indices])
        np.multiply(wk[self.output_indices], self._coeff[cdir],
                    wk[self.output_indices])
        np.abs(wk[self.output_indices], wk[self.output_indices])
        np.add(wk[self.output_indices], result[self.output_indices],
               result[self.output_indices])
        return result

if __name__ == '__main__':
    N = np.asarray((16,16,16))
    A = np.random.rand(*N)
    dx      = 1.0/(N-1)
    indices = np.ndindex(*N)

    diff = FDkCp(2,2,step=dx,indices=indices)
    diff(A,0,A)
