"""Test interpolation schemes
"""
from hysop.numerics.interpolation.interpolation import Linear
from hysop.tools.numpywrappers import npw
from hysop import Box, Discretization
import numpy as np
cos = np.cos

disc = [None, ] * 3
g = 1
nx = 33
ny = 69
nz = 129
disc[0] = Discretization([nx, ], [g, ])
disc[1] = Discretization([nx, ny], [g, g])
disc[2] = Discretization([nx, ny, nz], [g, g, g])
doms = [None, ] * 3
topos = [None, ] * 3
rworks = [None, ] * 3
iworks = [None, ] * 3
for case in xrange(3):
    dim = case + 1
    doms[case] = Box(origin=[0.] * dim, length=[4. * np.pi, ] * dim)
    topos[case] = doms[case].create_topology(disc[case])
    refsize = (np.prod(topos[case].mesh.compute_resolution) + 324, )
    iworks[case] = [npw.int_zeros(refsize)]
    rworks[case] = [npw.zeros(refsize)]
dx = topos[2].mesh.space_step


def run_interp(dimension, velo=0., with_work=False):
    """Interpolate a field and compare with reference
    (cosine function), for a given domain dimension
    """
    topo = topos[dimension - 1]
    rwork, iwork = None, None
    if with_work:
        rwork = rworks[dimension - 1]
        iwork = iworks[dimension - 1]
    ind = topo.mesh.compute_index
    field = npw.zeros(topo.mesh.local_resolution)
    for direction in xrange(topo.domain.dim):
        # lref = doms[dimension - 1].length[direction]
        h = topo.mesh.space_step[direction]
        x = topo.mesh.coords[direction]
        field[...] = np.cos(x)
        # select a set of points which are advected randomly and interpolated
        # --> 1 point over two from the original mesh
        shape = list(topo.mesh.local_resolution)
        coords = topo.mesh.coords[direction]
        ic = [slice(0, coords.shape[i]) for i in xrange(topo.domain.dim)]
        ic[direction] = slice(0, topo.mesh.local_resolution[direction] - 3, 2)
        shape[direction] = topo.mesh.coords[direction][ic].shape[direction]
        positions = npw.random(shape) * velo
        positions[...] += topo.mesh.coords[direction][ic]
        interp = Linear(field, direction=direction, topo_source=topo,
                        rwork=rwork, iwork=iwork)
        result = [npw.zeros_like(positions)]
        result = interp(0., [positions], result)
        reference = npw.zeros_like(result[0])
        reference[...] = np.cos(positions)
        tol = 0.5 * h ** 2
        assert np.allclose(result[0][ind], reference[ind], rtol=tol)


def test_linear_interpolation_1d_velocity_null():
    """Linear interpolation on 1d domain"""
    run_interp(1)


def test_linear_interpolation_1d():
    """Linear interpolation on 1d domain"""
    run_interp(1, 0.5 * dx[0])


def test_linear_interpolation_2d_velocity_null():
    """Linear interpolation on 2d domain"""
    run_interp(2)


def test_linear_interpolation_2d():
    """Linear interpolation on 2d domain"""
    run_interp(2, 0.5 * dx[1])


def test_linear_interpolation_3d_velocity_null():
    """Linear interpolation on 3d domain"""
    run_interp(3)


def test_linear_interpolation_3d():
    """Linear interpolation on 3d domain"""
    run_interp(3, 0.5 * dx[2])


def test_linear_interpolation_1d_wk_velocity_null():
    """Linear interpolation on 1d domain, external work arrays"""
    run_interp(1, 0., True)


def test_linear_interpolation_1d_wk():
    """Linear interpolation on 1d domain, external work arrays"""
    run_interp(1, 0.5 * dx[0], True)


def test_linear_interpolation_2d_wk_velocity_null():
    """Linear interpolation on 2d domain, external work arrays"""
    run_interp(2, 0., True)


def test_linear_interpolation_2d_wk():
    """Linear interpolation on 2d domain, external work arrays"""
    run_interp(2, 0.5 * dx[1], True)


def test_linear_interpolation_3d_wk_velocity_null():
    """Linear interpolation on 3d domain, external work arrays"""
    run_interp(3, 0., True)


def test_linear_interpolation_3d_wk():
    """Linear interpolation on 3d domain, external work arrays"""
    run_interp(3, 0.5 * dx[2], True)
