"""Interface to set a user-defined operator
"""
from hysop.operator.computational import Computational
from hysop.operator.discrete.custom import Custom as CO
from hysop.operator.continuous import opsetup


class Custom(Computational):
    """User-defined generic operator
    """
    def __init__(self, function, in_fields, out_fields=None,
                 diagnostics_shape=None, **kwds):
        """

        Parameters
        ----------
        in_fields: list of :class:`~hysop.fields.discrete_field.DiscreteField`
             input fields args. for function, see notes below
        out_fields: list of :class:`~hysop.fields.discrete_field.DiscreteField`
             output fields args for function, see notes below
        function: python function
             a user defined function, called by this op.apply method.
        diagnostics_shape: tuple, optional
             shape of the data expected to be written into file.

        Notes
        -----
        A function is used to set the behavior of the current operator,
        during apply call.
        This function must look like::

            def some_func(simulation, in_fields, out_fields, diagnostics=None):
                # do things ...

        and compute out_fields values and optionnaly some diagnostics.
        See :ref:`custom`
        """
        super(Custom, self).__init__(**kwds)
        self.function = function
        self.input = in_fields
        if out_fields is not None:
            self.output = out_fields
        self._diagnostics_shape = diagnostics_shape

    def discretize(self):
        super(Custom, self)._standard_discretize()

    @opsetup
    def setup(self, rwork=None, iwork=None):
        if not self._is_uptodate:
            self.discrete_op = CO(
                in_fields=[self.discrete_fields[f] for f in self.input],
                out_fields=[self.discrete_fields[f] for f in self.output],
                function=self.function,
                variables=self.discrete_fields.values())
            if self._diagnostics_shape is not None:
                assert isinstance(self._diagnostics_shape, tuple)
                assert len(self._diagnostics_shape) == 2
                self._set_io(self.function.__name__, self._diagnostics_shape)
                self.discrete_op.set_writer(self._writer)

            self._is_uptodate = True

    def get_work_properties(self):
        return {'rwork': None, 'iwork': None}
