# -*- coding: utf-8 -*-
"""Operator for diffusion problem.

See :ref:`diffusion` in HySoP user guide.


"""
from hysop.operator.computational import Computational
from hysop.operator.discrete.diffusion_fft import DiffusionFFT,\
    CurlAndDiffusionFFT
from hysop.constants import debug
from hysop.operator.continuous import opsetup
from hysop.methods import SpaceDiscretization, GhostUpdate
from hysop import __FFTW_ENABLED__, __GPU_ENABLED__
if __GPU_ENABLED__:
    from hysop.backend.device.opencl.gpu_diffusion import GPUDiffusion


class Diffusion(Computational):
    """Diffusion of a field.
    """

    _authorized_methods = ['fftw', 'on_gpu']

    @debug
    def __init__(self, viscosity, vorticity, **kwds):
        """Diffusion operator.
        See :ref:`diffusion` in HySoP user guide.

        Parameters
        ----------
        viscosity : double
             constant viscosity value
        vorticity : :class:`~hysop.fields.continuous_field.Field`
             vorticity field, in/out parameter.
        kwds : base class parameters.
        """
        assert 'variables' not in kwds, 'variables parameter is useless.'
        super(Diffusion, self).__init__(variables=[vorticity], **kwds)
        if 'method' not in kwds:
            import hysop.default_methods as default
            self.method = default.DIFFUSION
        else:
            self.method = kwds.pop('method')
        
        msg = 'Diffusion : unknown method for space discretization'
        assert self.method[SpaceDiscretization] in self._authorized_methods,\
               msg

        msg = 'Diffusion : on_gpu resolution is not available on your system.'
        if self.method[SpaceDiscretization] is 'on_gpu':
            assert __GPU_ENABLED__, msg

        # input/output field, solution of the problem
        self.vorticity = vorticity
        # viscosity
        self.viscosity = viscosity
        # kwds required for gpu discrete operator
        self.kwds = kwds.copy()
        if 'discretization' in self.kwds:
            self.kwds.pop('discretization')
        self.input = [self.vorticity]
        self.output = [self.vorticity]

    def discretize(self):
        if self.method[SpaceDiscretization] is 'fftw' and __FFTW_ENABLED__:
            super(Diffusion, self)._fftw_discretize()
        elif self.method[SpaceDiscretization] is 'on_gpu':
            super(Diffusion, self)._standard_discretize()
        else:
            raise AttributeError("Method not yet implemented.")

    @debug
    @opsetup
    def setup(self, rwork=None, iwork=None):
        if self.method[SpaceDiscretization] is 'fftw':
            dop = DiffusionFFT
        elif self.method[SpaceDiscretization] is 'on_gpu':
            dop = GPUDiffusion

        self.discrete_op = dop(
            viscosity=self.viscosity,
            vorticity=self.discrete_fields[self.vorticity],
            method=self.method, **self.kwds)
        self._is_uptodate = True

    def get_work_properties(self):
        return {'rwork': None, 'iwork': None}


class CurlAndDiffusion(Computational):
    """Solve curl of velocity and its diffusion in one shot
    (in Fourier domain).
    """

    _authorized_methods = ['fftw']

    @debug
    def __init__(self, viscosity, velocity, vorticity, **kwds):
        """Diffusion operator.
        See :ref:`diffusion` in HySoP user guide.

        Parameters
        ----------
        viscosity : double
             constant viscosity value
        velocity : :class:`~hysop.fields.continuous_field.Field
             vorticity field, in/out parameter.
        vorticity : :class:`~hysop.fields.continuous_field.Field
             vorticity field, in/out parameter.
        kwds : base class parameters.

        Notes:
        * vorticity parameter is optional since the field
        can be provided in the usual way for operators using
        variables parameter::

            op = Diffusion(variables={vorticity: topo}, ...)

        """
        self.viscosity = viscosity
        self.velocity = velocity
        self.vorticity = vorticity
        msg = 'fftw required for CurlAndDiffusion. '
        msg += 'Try to recompile with WITH_FFTW=ON'
        assert __FFTW_ENABLED__, msg
        msg = 'CurlAndDiffusion : unknown method for space discretization'
        self.method = {SpaceDiscretization: 'fftw', GhostUpdate: True}
        self.input = [self.velocity]
        self.output = [self.vorticity]
        assert 'variables' not in kwds, 'variables parameter is useless.'
        super(CurlAndDiffusion, self).__init__(
            variables=[velocity, vorticity], **kwds)

    def discretize(self):
        super(CurlAndDiffusion, self)._fftw_discretize()

    @debug
    @opsetup
    def setup(self, rwork=None, iwork=None):
        self.discrete_op = CurlAndDiffusionFFT(
            viscosity=self.viscosity,
            velocity=self.discrete_fields[self.velocity],
            vorticity=self.discrete_fields[self.vorticity],
            method=self.method)
        self._is_uptodate = True

    def get_work_properties(self):
        return {'rwork': None, 'iwork': None}
