"""Advection solver, particular method, pure-python version

"""
from hysop.constants import debug, WITH_GUESS, HYSOP_REAL, HYSOP_INTEGER, EPS
from hysop.methods import TimeIntegrator, Interpolation, Remesh
from hysop.operator.discrete.discrete import DiscreteOperator
from hysop.default_methods import ADVECTION
import numpy as np
from hysop.numerics.remeshing import Remeshing
from hysop.tools.profiler import profile
from hysop.tools.misc import WorkSpaceTools
from hysop.tools.numpywrappers import npw


class ParticleAdvection(DiscreteOperator):
    """
    Particular method for advection of a list of fields in a given direction.
    """

    @debug
    def __init__(self, velocity, fields_on_grid, direction, **kwds):
        """
        Parameters
        ----------
        velocity : :class:`~hysop.fields.discrete_field.DiscreteField`
            advection velocity (discretized)
        fields_on_grid : list of :class:`~hysop.fields.discrete_field.DiscreteField`
            discrete field(s) to be advected
        direction : int
            advection direction
        kwds : extra parameters for base class

        """
        msg = 'Scales advection init : variables parameter is useless.'
        msg += 'See user guide for details on the proper'
        msg += ' way to build the operator.'
        assert 'variables' not in kwds, msg
        # Advection velocity
        self.velocity = velocity

        # set variables list ...
        variables = [self.velocity]
        if not isinstance(fields_on_grid, list):
            self.fields_on_grid = [fields_on_grid]
        else:
            self.fields_on_grid = fields_on_grid
        variables += self.fields_on_grid

        if 'method' not in kwds:
            kwds['method'] = ADVECTION
        # number of components of the RHS (see time integrator)
        self._rhs_size = 1
        # dictionnary of internal buffers used
        # to save fields values on particles
        # --> link to _rwork
        self.fields_on_part = None
        # buffers for numerical methods
        # --> just link to _rwork/_iwork components
        self._rw_interp = None
        self._iw_interp = None
        self._rw_integ = None
        self._rw_remesh = None
        self._iw_remesh = None
        # buffer used to save particles position
        # --> link to _rwork
        self.part_positions = None

        super(ParticleAdvection, self).__init__(variables=variables, **kwds)

        self.input = self.variables
        self.output = [df for df in self.variables if df is not self.velocity]
        self.direction = direction
        # time integrator
        self.num_advec = None
        # remeshing method
        self.num_remesh = None
        # build and configure numerical methods
        self._configure_numerical_methods()
        # minimal value required for advected field
        # on grid point to add a particle.
        # If None, particles everywhere.
        self._threshold = None
        self._init_particles_on_grid = self._init_particles_everywhere
        if self._threshold is not None:
            self._init_particles_on_grid = self._init_particles_with_threshold

    def _configure_numerical_methods(self):
        """Function to set the numerical method for python operator and link them
        to the proper working arrays.

        Warning : must be called after _set_work_arrays
        """

        # ---  numerical interpolation operator ---
        # discrete field that will be interpolated
        vd = self.velocity.data[self.direction]
        # first field topology as reference
        topo_fields = self.fields_on_grid[0].topology
        # velocity 'grid'
        topo_velo = self.velocity.topology
        # -- Interpolation scheme --
        # to interpolate velocity from topo_velo
        # to the grid on which particles are initialized.
        num_interpolate = \
            self.method[Interpolation](vd, direction=self.direction,
                                       topo_source=topo_velo,
                                       topo_target=topo_fields,
                                       rwork=self._rw_interp,
                                       iwork=self._iw_interp)

        # -- time integration --
        # advection is performed only on 'computational points',
        # not on ghosts
        ic = topo_fields.mesh.compute_index
        self.num_advec = self.method[TimeIntegrator](
            self._rhs_size, rwork=self._rw_integ,
            f=num_interpolate,
            topo=topo_fields,
            optim=WITH_GUESS,
            indices=ic)

        # -- remesh --
        self.num_remesh = Remeshing(self.method[Remesh],
                                    topo_source=topo_fields,
                                    direction=self.direction,
                                    rwork=self._rw_remesh,
                                    iwork=self._iw_remesh)

    def _set_work_arrays(self, rwork=None, iwork=None):

        # Find number and shape of required work arrays

        # topologies for advected field(s) grid
        topo_fields = self.fields_on_grid[0].topology
        # -- work arrays properties for interpolation --
        # work arrays shape depends on the targeted grid, i.e. topo_fields.
        interp_wp = self.method[Interpolation].get_work_properties(topo_fields)
        interp_iwork_length = len(interp_wp['iwork'])
        interp_rwork_length = len(interp_wp['rwork'])

        # -- work arrays properties for time-integrator --
        # depends on topo_fields.
        ti_wp = self.method[TimeIntegrator].get_work_properties(
            self._rhs_size, topo_fields)
        ti_rwork_length = len(ti_wp['rwork'])
        # -- work arrays properties for remeshing scheme --
        # depends on topo_fields.
        remesh_wp = Remeshing.get_work_properties(topo_fields)
        remesh_iwork_length = len(remesh_wp['iwork'])
        remesh_rwork_length = len(remesh_wp['rwork'])

        # -- Find out the number of required work arrays --
        # Interpolation and odesolver work arrays must be different
        # but after interpolation/advection, work arrays can be reused
        # by remeshing. So the number of required work arrays is
        # equal to
        # max(nb works_for_time_integrator + nb work_for_interpolation,
        #     nb works_for_remesh)
        rwork_length = max(remesh_rwork_length,
                           ti_rwork_length + interp_rwork_length)
        iwork_length = max(interp_iwork_length, remesh_iwork_length)

        # -- Finally, particles positions and fields on particles
        # values will also be in work array.
        ppos_index = rwork_length
        # max number of components for advected fields
        nbc = max([f.nb_components for f in self.fields_on_grid])
        rwork_length += nbc
        rwork_length += 1  # work array for positions

        # -- rwork organisation --

        # rw = [ | rw_interp | rw_time_int | part_pos | fields_on_part |]
        #        |        rw_remesh        |      ...                  |]

        #  work array shape depends on the time integrator
        #  interpolation scheme and remeshing scheme

        # check and/or allocate work arrays according to properties above
        # reshape stuff is done inside numerical methods.
        # Here we just allocate flat memory.
        subshape_field = tuple(topo_fields.mesh.compute_resolution)
        subsize_interp_ti = interp_wp['rwork'] + ti_wp['rwork']
        subsize_remesh = remesh_wp['rwork']
        # list of shapes for rwork arrays (!flat arrays!)
        subsize_rw = max(subsize_interp_ti, subsize_remesh)
        # list of shapes for iwork arrays (!flat arrays!)
        subsize_iw = max(interp_wp['iwork'], remesh_wp['iwork'])
        # one array for particles positions
        # and arrays for fields on particles
        for i in xrange(nbc + 1):
            subsize_rw.append((np.prod(subshape_field), ))

        # Build iwork and rwork lists
        self._rwork = WorkSpaceTools.check_work_array(rwork_length, subsize_rw,
                                                      rwork, HYSOP_REAL)
        self._iwork = WorkSpaceTools.check_work_array(iwork_length, subsize_iw,
                                                      iwork, HYSOP_INTEGER)

        # -- set connections for remesh, ti and interp methods --
        # interpolation
        self._rw_interp = [self._rwork[i] for i in xrange(interp_rwork_length)]
        self._iw_interp = [self._iwork[i] for i in xrange(interp_iwork_length)]
        assert np.asarray([npw.arrays_share_data(self._rw_interp[i],
                                                 self._rwork[i])
                           for i in xrange(interp_rwork_length)]).all()

        # time integration
        self._rw_integ = [self._rwork[i]
                          for i in xrange(interp_rwork_length,
                                          interp_rwork_length +
                                          ti_rwork_length)]
        # remesh
        self._rw_remesh = [self._rwork[i] for i in xrange(remesh_rwork_length)]
        self._iw_remesh = [self._iwork[i] for i in xrange(remesh_iwork_length)]

        # --- links to work for particles positions buffers ---
        self._work_part_pos = ppos_index
        self.part_positions = WorkSpaceTools.check_work_array(
            1, subshape_field, [self._rwork[ppos_index]])
        assert npw.arrays_share_data(self.part_positions[0],
                                     self._rwork[ppos_index])

        # --- links to buffers for fields on particles ---
        ppos_index += 1
        self.fields_on_part = WorkSpaceTools.check_work_array(
            nbc, subshape_field,
            [self._rwork[i] for i in xrange(ppos_index, ppos_index + nbc)])
        assert np.asarray(
            [npw.arrays_share_data(self.fields_on_part[i],
                                   self._rwork[ppos_index + i])
             for i in xrange(nbc)]).all()

    @debug
    @profile
    def apply(self, simulation=None):
            #, dt_coeff=1., split_id=0, old_dir=0):
        """
        Advection algorithm:
        - initialize particles and fields with their values on the grid.
          --> x_p, f_p
        - compute particle positions in splitting direction,
          --> solve dx_p / dt = v_p
        - remesh fields from particles to grid
          --> f_g
        """
        assert simulation is not None, \
            'Simulation parameter is missing.'
        t, dt = simulation.time, simulation.sub_step

        # -- Initialize particles on the grid --
        self._init_particles_on_grid()

        # -- Advect particles --
        # RHS of odesolver must be computed and save in odesolver._rwork[0]
        # Notes :
        # - ghost-synchro of the rhs (if needed) is done by the odesolver
        # - interpolation of velocity field from grid to particles is done
        # 'inside' the ode solver (rhs = interpolation scheme)
        self.num_advec.rwork[0][...] = self.velocity.data[self.direction][...]
        self.part_positions = self.num_advec(
            t, self.part_positions, dt, result=self.part_positions)

        # -- Remesh --
        for fg in self.fields_on_grid:
            # Initialize fields on particles with fields on grid values --
            for d in xrange(fg.nb_components):
                self.fields_on_part[d][...] = fg[d][...]
            # Remesh all components of the field
            fg = self.num_remesh(ppos=self.part_positions,
                                 pscal=self.fields_on_part, result=fg)

    def _init_particles_everywhere(self):
        """Initialize particles on grid points
        """
        # indices of computation points (i.e. no ghosts)
        topo = self.fields_on_grid[0].topology
        ic = topo.mesh.compute_index
        self.part_positions[0][...] = \
            topo.mesh.coords[self.direction][ic[self.direction]] + 10 * EPS
        # remark : we shift particles with EPS to avoid
        # the case where, because of floating point precision
        # xp (particle position) < xg (grid position),
        # with xp - xg very small. If this happens,
        # left point of xp will be xg - 1 rather than xg
        # which is not what is expected.

    def _init_particles_with_threshold(self):
        """"Initialize particles on grid points where
        advected field value is greater than a given threshold
        """
        raise AttributeError("Not yet implemented")
        # first field is used as reference to check threshold
        ref_field = self.fields_on_grid[0].data
        topo = self.fields_on_grid[0].topology
        ic = topo.mesh.compute_index
        self._rwork[0][...] = 0.
        for i in xrange(len(ref_field)):
            self._rwork[0][...] += ref_field[i][ic] ** 2
        self._rwork[0][...] = np.sqrt(self._rwork[0])
        ipart = np.where(self._rwork < self._threshold)
        part_shape = ref_field[0][ipart].shape
        self.part_positions = WorkSpaceTools.check_work_array(
            1, part_shape, self._rwork[self._work_part_pos])
        self.part_positions[0][...] = topo.mesh.coords[self.direction][ipart]
