# -*- coding: utf-8 -*-
"""Discretization of the stretching operator.

Formulations:

* :class:`~hysop.operator.discrete.stretching.Conservative`
* :class:`~hysop.operator.discrete.stretching.GradUW`

"""

from hysop.constants import debug, WITH_GUESS
from hysop.methods import TimeIntegrator, SpaceDiscretization
from hysop.operator.discrete.discrete import DiscreteOperator
from hysop.numerics.differential_operations import GradVxW, DivWV
from hysop.tools.numpywrappers import npw
from hysop.numerics.update_ghosts import UpdateGhosts
from hysop.tools.profiler import profile
from hysop.tools.misc import WorkSpaceTools
from abc import ABCMeta, abstractmethod
import math
import numpy as np
ceil = math.ceil


class Stretching(DiscreteOperator):
    """Abstract interface to stretching discrete operators.
    Check user guide to see the available formulations.
    """

    __metaclass__ = ABCMeta

    @debug
    def __init__(self, velocity, vorticity, formulation,
                 rhs, extra_ghosts_fields=None, **kwds):
        """Abstract interface for stretching operator

        Parameters
        ----------
        velocity, vorticity : :class:`~hysop.fields.discrete_field.DiscreteField`
        formulation : one of the discrete stretching classes.
        rhs : python function
            right-hand side for the time integrator
        extra_ghosts_fields : int, optional
            number of additional numpy arrays to be synchronized (mpi),
            default = None, i.e. synchro of vorticity and velocity components.
        **kwds : extra parameters for base class

        """
        # velocity discrete field
        self.velocity = velocity
        # vorticity discrete field
        self.vorticity = vorticity
        # Formulation for stretching (divWV or GradVxW)
        self.formulation = formulation

        if 'method' not in kwds:
            import hysop.default_methods as default
            kwds['method'] = default.STRETCHING
        # Work vector used by time-integrator
        self._ti_work = None
        # Work vector used by numerical diff operator.
        self._str_work = None
        # all works vectors are set in base class init,
        # thanks to _set_work_arrays function.
        super(Stretching, self).__init__(variables=[self.velocity,
                                                    self.vorticity], **kwds)

        self.input = self.variables
        self.output = [self.vorticity]
        # \todo multiresolution case
        assert self.velocity.topology.mesh == self.vorticity.topology.mesh,\
            'Multiresolution case not yet implemented.'

        # Number of components of the operator (result)
        self.nb_components = 3  # Stretching only in 3D and for vector fields.

        # prepare ghost points synchro for velocity and vorticity
        nb_ghosts_fields = self.velocity.nb_components + \
            self.vorticity.nb_components
        if extra_ghosts_fields is not None:
            nb_ghosts_fields += extra_ghosts_fields
        self._synchronize = UpdateGhosts(self.velocity.topology,
                                         nb_ghosts_fields)

        # A function to compute the gradient of the velocity.
        # Work vector is provided in input.
        self.strFunc = \
            self.formulation(topo=self.velocity.topology,
                             work=self._str_work,
                             method=self.method[SpaceDiscretization])
        # Time integrator
        self.timeIntegrator = \
            self.method[TimeIntegrator](self.nb_components,
                                        self.velocity.topology,
                                        rwork=self._ti_work,
                                        f=rhs,
                                        optim=WITH_GUESS)
        # In self._integrate ti_work is used as in/out parameters in the rhs,
        # of the time-integrator so it needs a reshape.
        # Warning : this is done without any memory realloc.
        for i in xrange(len(self._ti_work)):
            self._ti_work[i] = self.timeIntegrator.rwork[i]

    def _set_work_arrays(self, rwork=None, iwork=None):

        ti = self.method[TimeIntegrator]
        topo = self.velocity.topology
        nbc = self.velocity.nb_components
        # properties of working arrays required for time-integrator
        wk_prop = ti.get_work_properties(nbc, topo)['rwork']
        lenti = len(wk_prop)
        # properties of working arrays required for differential operations,
        wk_prop += self.formulation.get_work_properties(topo)['rwork']
        subshape = np.prod(topo.mesh.local_resolution)
        self._rwork = WorkSpaceTools.check_work_array(len(wk_prop),
                                                      subshape, rwork)
        self._ti_work = self._rwork[:lenti]
        self._str_work = self._rwork[lenti:]

    @profile
    def update_ghosts(self):
        """Ghost points synchronization
        """
        self._synchronize(self.velocity.data + self.vorticity.data)

    def apply(self, simulation=None):
        """Solve stretching equation and update vorticity
        """
        assert simulation is not None, \
            "Missing simulation value for computation."

        # time step
        dt = simulation.time_step
        # current time
        t = simulation.time

        # Synchronize ghost points of velocity
        self._synchronize(self.velocity.data + self.vorticity.data)
        self._compute(dt, t)

    @abstractmethod
    def _compute(self, dt, t):
        """Abstract interface to local stretching resolution
        """

    @profile
    def _integrate(self, dt, t):
        """Apply time integrator
        """
        # Init workspace with a first evaluation of the
        # rhs of the integrator
        self._ti_work[:self.nb_components] = \
            self.timeIntegrator.f(t, self.vorticity.data,
                                  self._ti_work[:self.nb_components])
        # perform integration and save result in-place
        self.vorticity.data = self.timeIntegrator(t, self.vorticity.data, dt,
                                                  result=self.vorticity.data)


class Conservative(Stretching):
    """Conservative formulation
    """

    @profile
    def __init__(self, **kwds):

        # Right-hand side for time integration
        def rhs(t, y, result):
            """rhs used in time integrator
            """
            return self.strFunc(y, self.velocity.data, result)

        super(Conservative, self).__init__(
            formulation=DivWV, rhs=rhs, **kwds)

    @profile
    def _compute(self, dt, t):
        # No subcycling for this formulation
        self._integrate(dt, t)


class GradUW(Stretching):
    """GradUW formulation
    """

    def __init__(self, **kwds):

        def rhs(t, y, result):
            """rhs used in time integrator
            """
            result, self.diagnostics =\
                self.strFunc(self.velocity.data, y, result, self.diagnostics)
            return result

        super(GradUW, self).__init__(formulation=GradVxW, rhs=rhs, **kwds)

        # stability constant
        # Depends on time integration method
        self.cststretch = self.method[TimeIntegrator].stability_coeff()
        # a vector to save diagnostics computed from GradVxW (max div ...)
        self.diagnostics = npw.zeros(2)
        self.diagnostics[1] = self.cststretch

    @debug
    @profile
    def _compute(self, dt, t):
        # Compute the number of required subcycles
        ndt, subdt = self._check_stability(dt)
        assert sum(subdt) == dt

        for i in xrange(ndt):
            self._integrate(subdt[i], t)

    def _check_stability(self, dt):
        """Computes a stability condition depending on some
        diagnostics (from GradVxW)

        :param dt: current time step

        Returns
        --------
        nb_cylces : int
            the number of required subcycles and
        subdt : array of float
            the subcycles time-step.
        """
        dt_stab = min(dt, self.cststretch / self.diagnostics[1])
        nb_cycles = int(ceil(dt / dt_stab))
        subdt = npw.zeros((nb_cycles))
        subdt[:] = dt_stab
        subdt[-1] = dt - (nb_cycles - 1) * dt_stab
        return nb_cycles, subdt


class StretchingLinearized(Stretching):
    """Conservative formulation of the linearized stretching.
    """

    def __init__(self, vorticity_BF, usual_op, **kwds):
        # vorticity of the base flow (steady solution)
        self.vorticity_BF = vorticity_BF
        # prepare ghost points synchro for vorticity_BF
        self._synchronize_vort_BF = \
            UpdateGhosts(self.vorticity_BF.topology,
                         self.vorticity_BF.nb_components)
        self.usual_op = usual_op
        # boolean used  to switch between two rhs forms:
        # either rhs(t, y) = div(y:u) (if true)
        # or rhs(t,y) = div(y:w_bf) (if false)
        self._divwu = True

        def rhs(t, y, result):
            """rhs used in time integrator
            """
            if self._divwu:
                result = self.strFunc(y, self.velocity.data, result)
            else:
                result = self.strFunc(y, self.vorticity_BF.data, result)
            return result

        super(StretchingLinearized, self).__init__(
            formulation=DivWV, rhs=rhs,
            extra_ghosts_fields=self.vorticity_BF.nb_components,
            **kwds)

    def _compute(self, dt, t):
        # - Call time integrator (1st term over 3) -
        # Init workspace with a first evaluation of the div(wb:u') term in the
        # rhs of the integrator
        self._divwu = True
        self._ti_work[:self.nb_components] = \
            self.timeIntegrator.f(t, self.vorticity_BF.data,
                                  self._ti_work[:self.nb_components])
        # perform integration and save result in-place
        self.vorticity.data = self.timeIntegrator(t, self.vorticity.data, dt,
                                                  result=self.vorticity.data)
        # - Call time integrator (2nd term over 3) -
        # Init workspace with a first evaluation of the div(u':wb) term in the
        # rhs of the integrator
        self._divwu = False
        self._ti_work[:self.nb_components] = \
            self.timeIntegrator.f(t, self.velocity.data,
                                  self._ti_work[:self.nb_components])
        # perform integration and save result in-place
        self.vorticity.data = self.timeIntegrator(t, self.vorticity.data, dt,
                                                  result=self.vorticity.data)

    def apply(self, simulation=None):
        assert simulation is not None, \
            "Missing simulation value for computation."

        # time step
        dt = simulation.time_step
        # current time
        t = simulation.time

        # Synchronize ghost points
        self._synchronize(self.velocity.data + self.vorticity.data +
                          self.vorticity_BF.data)
        # Compute the 2 first "stretching" terms (div(wb:u') and div(u':wb))
        # and update vorticity for each of them
        self._compute(dt, t)
        # Compute the 3rd stretching term (div(w':ub)) and update vorticity
        self.usual_op.apply(simulation)
