"""Filter values between grids of different resolution.

"""
from hysop.constants import debug
from hysop.operator.continuous import opsetup
from hysop.operator.computational import Computational
import hysop.default_methods as default
from hysop.methods import Support


class MultiresolutionFilter(Computational):
    """Interpolation from fine grid to coarse grid

    """

    @debug
    def __init__(self, d_in, d_out, **kwds):
        """
        Parameters
        ----------
        d_in, d_out : :class:`hysop.topology.topology.CartesianTopology`
            or :class:`tools.parameters.Discretization`
            data distribution for source (in) and target (out) grids.
        kwds : base class parameters

        """
        if 'method' not in kwds:
            kwds['method'] = default.MULTIRESOLUTION_FILTER
        super(MultiresolutionFilter, self).__init__(**kwds)
        self.d_in, self.d_out = d_in, d_out
        self.input = self.variables
        self.output = self.variables
        self._df_in = []
        self._df_out = []

    def discretize(self):
        super(MultiresolutionFilter, self)._standard_discretize()
        topo_in = self._build_topo(self.d_in, 0)
        topo_out = self._build_topo(self.d_out, 0)
        for v in self.variables:
            self._df_in.append(v.discretize(topo_in))
            self._df_out.append(v.discretize(topo_out))

    @opsetup
    def setup(self, rwork=None, iwork=None):
        if Support in self.method.keys() and \
           self.method[Support].find('gpu') >= 0:
            from hysop.backend.device.opencl.gpu_multiresolution_filter \
                import GPUFilterFineToCoarse as discreteFilter
        else:
            from hysop.operator.discrete.multiresolution_filter \
                import FilterFineToCoarse as discreteFilter
        self.discrete_op = discreteFilter(
            field_in=self._df_in, field_out=self._df_out,
            method=self.method, rwork=rwork, iwork=iwork)
        self._is_uptodate = True

    def get_work_properties(self):
        return {'rwork': None, 'iwork': None}
