# -*- coding: utf-8 -*-
"""Operator to solve Poisson problem.

See :ref:`poisson` in HySoP user guide.
"""
from hysop.operator.computational import Computational
from hysop.operator.discrete.poisson_fft import PoissonFFT
from hysop.constants import debug
from hysop.operator.velocity_correction import VelocityCorrection
from hysop.operator.reprojection import Reprojection
from hysop.methods import SpaceDiscretization, Formulation
from hysop.operator.continuous import opsetup
from hysop import __FFTW_ENABLED__


class Poisson(Computational):
    """Solve Poisson problem (in: vorticity, out velocity)
    for incompressible flow.
    """

    _authorized_methods = ['fftw']

    @debug
    def __init__(self, output_field, input_field, flowrate=None,
                 projection=None, **kwds):
        """Poisson operator, incompressible flow.

        Parameters
        ----------
        output_field : :class:`~hysop.fields.continuous_field.Field
            solution field
        input_field  : :class:`~hysop.fields.continuous_field.Field`
            right-hand side
        flowrate: :class:`~hysop.fields.variable_parameter.VariableParameter`
           or double, optional
            flow rate value through input surface (normal to xdir),
            used to calibrate solution, default=0.
        projection : double or tuple, optional
             projection criterion, see notes below.
        kwds : base class parameters.


        Notes:
        * projection might be:
           * None: no projection
           * the value of the frequency of reprojection (constant)
           * a tuple (frequency, threshold). In that case, a criterion
           depending on the input_field will be computed at each time step,
           and if criterion > threshold., then frequency projection is active.
        * About method parameter:
           - SpaceDiscretization == fftw
           - Formulation = 'velocity' or 'pressure'
           velocity : laplacian(phi) = -w and v = nabla X psi, in = vorticity, out = velo
           pressure : laplacian(p) = -nabla.(u.nabla u, in = velo, out = pressure
        """
        # Warning : for fftw all variables must have
        # the same resolution.
        assert 'variables' not in kwds, 'variables parameter is useless.'
        super(Poisson, self).__init__(variables=[output_field, input_field],
                                      **kwds)
        # solution of the problem
        self.output_field = output_field
        # -(right-hand side)
        self.input_field = input_field
        if self.method is None:
            import hysop.default_methods as default
            self.method = default.POISSON
        if self.method[SpaceDiscretization] is 'fftw':
            assert __FFTW_ENABLED__
        msg = 'Poisson : unknown method for space discretization'
        assert self.method[SpaceDiscretization] in self._authorized_methods,\
            msg

        # Set Poisson equation formulation :
        # Velo Poisson eq or Pressure Poisson eq
        self.formulation = None
        if self.method[Formulation] is not 'velocity':
            self.formulation = self.method[Formulation]

        self.input = [self.input_field]
        self.output = [self.output_field]
        # Enable correction if required
        if flowrate is not None:
            self.withCorrection = True
            self._flowrate = flowrate
        else:
            self.withCorrection = False
        self.correction = None
        self.projection = projection
        self._config = kwds

        if projection is not None:
            self.output.append(self.input_field)

    def discretize(self):
        # Poisson solver based on fftw
        if self.method[SpaceDiscretization] is 'fftw':
            super(Poisson, self)._fftw_discretize()
            # prepare correction and projection, if needed
            if self.withCorrection:
                toporef = self.discrete_fields[self.output_field].topology
                if 'discretization' in self._config:
                    self._config['discretization'] = toporef
                self.correction = VelocityCorrection(
                    self.output_field, self.input_field,
                    req_flowrate=self._flowrate, **self._config)
                self.correction.discretize()

                if isinstance(self.projection, tuple):
                    freq = self.projection[0]
                    threshold = self.projection[1]
                    self.projection = Reprojection(self.input_field,
                                                   threshold, freq,
                                                   **self._config)
                    self.projection.discretize()

    @debug
    @opsetup
    def setup(self, rwork=None, iwork=None):
        # Activate correction, if required
        if self.withCorrection:
            self.correction.setup()
            cd = self.correction.discrete_op
        else:
            cd = None

        # Activate projection, if required
        if isinstance(self.projection, Reprojection):
            # Projection frequency is updated at each
            # time step, and depends on the input_field
            self.projection.setup(rwork=rwork)
            projection_discr = self.projection.discrete_op
        else:
            projection_discr = self.projection

        self.discrete_op = PoissonFFT(self.discrete_fields[self.output_field],
                                      self.discrete_fields[self.input_field],
                                      correction=cd,
                                      rwork=rwork, iwork=iwork,
                                      projection=projection_discr,
                                      formulation=self.formulation)

        self._is_uptodate = True

    def get_work_properties(self):
        return {'rwork': None, 'iwork': None}
