"""Setup for data transfer/redistribution between topologies or operators

`.. currentmodule : hysop.operator.redistribute

* :class:`~RedistributeIntra` for topologies/operators defined
  inside the same mpi communicator
* :class:`~RedistributeInter` for topologies/operators defined
  on two different mpi communicator
* :class:`~RedistributeOverlap` for topologies defined
  inside the same mpi parent communicator and
  with a different number of processes
* :class:`~Redistribute` abstract base class

"""

from hysop.operator.continuous import OperatorBase
from abc import ABCMeta, abstractmethod
from hysop.topology.cartesian_topology import CartesianTopology
from hysop.operator.computational import Computational
from hysop.operator.continuous import opsetup, opapply
from hysop.core.mpi.bridge import Bridge, BridgeOverlap, BridgeInter
from hysop.constants import DirectionLabels, debug


class Redistribute(OperatorBase):
    """Abstract interface to redistribute operators
    """

    __metaclass__ = ABCMeta

    def __init__(self, source, target, component=None,
                 run_till=None, **kwds):
        """
        Parameters
        ----------
        source, target: :class:`~hysop.topology.topology.CartesianTopology` or
         :class:`~hysop.operator.computational.Computational
            topologies or operators that own the source mesh and targeted mesh
        component: int
            which component of the field must be distributed (default = all)
        run_till: list of :class:`~hysop.operator.computational.Computational
            operators that must wait for the completion of this redistribute
            before any apply.

        """
        # Base class initialisation
        super(Redistribute, self).__init__(**kwds)

        # Object (may be an operator or a topology) which handles the
        # fields to be transfered
        self._source = source
        # Object (may an operator or a topology) which handles the fields
        # to be filled in from source.
        self._target = target

        self.component = component
        if component is None:
            # All components are considered
            self._range_components = lambda v: xrange(v.nb_components)
        else:
            # Only the given component is considered
            assert self.component >= 0, 'component value must be positive.'
            self._range_components = lambda v: (self.component)

        # Bridge between topology of source and topology of target
        self.bridge = None
        # True if some MPI operations are running for the current operator.
        self._has_requests = False
        # Which operator must wait for this one before
        # any computation
        # Exp : run_till = op1 means that op1 will
        # wait for the end of this operator before
        # op1 starts its apply.
        if run_till is None:
            run_till = []

        assert isinstance(run_till, list)
        self._run_till = run_till

    @abstractmethod
    def setup(self, rwork=None, iwork=None):
        """
        Check/set the list of variables to be distributed

        What must be set at setup?
        ---> the list of continuous variables to be distributed
        ---> the bridge (one for all variables, which means
        that all vars must have the same topology in source
        and the same topology in target.
        ---> the list of discrete variables for source and
        for target.
        """
        assert self.domain is not None
        for v in self.variables:
            assert v.domain is self.domain
        super(Redistribute, self).setup(rwork, iwork)

    def _check_operator(self, op):
        """ ensure op properties:
           * check if op is really a computational operator
           * discretize op
           * check if all required variables (if any) belong to op

        Parameters
        ----------
        op : :class:`~hysop.operator.computational.Computational
                :param:  op : a computational operator

        """
        assert isinstance(op, Computational)
        op.discretize()
        msg = 'The variables to be distributed '
        msg += 'do not belong to the input operator.'
        if len(self.variables) > 0:
            assert all(v in op.variables for v in self.variables), msg

    def _set_variables(self):
        """
        Check/set the list of variables proceed by the current operator.
        """
        # Set list of variables.
        # It depends on :
        # - the type of source/target : CartesianTopology, Computational or None
        # - the args variables : a list of variables or None
        # Possible cases:
        # - if source or target is None --> variables is required
        # - if source and target are CartesianTopology --> variables is required
        # - in all other cases, variables is optional.
        # If variables are not set at init,
        # they must be infered from source/target operators.
        has_var = len(self.variables) > 0
        vlist = (v for v in self.variables)

        if self._source is None or self._target is None:
            assert len(self.variables) > 0
            self.variables = [v for v in vlist]
        else:
            source_is_topo = isinstance(self._source, CartesianTopology)
            target_is_topo = isinstance(self._target, CartesianTopology)

            # both source and target are topologies. Variables required.
            if source_is_topo and target_is_topo:
                msg = 'Redistribute, a list of variables is required at init.'
                assert has_var, msg
                self.variables = [v for v in vlist]

            elif not source_is_topo and not target_is_topo:
                # both source and target are operators
                # --> intersection of their variables
                vsource = self._source.variables
                vtarget = self._target.variables
                if not has_var:
                    vlist = (v for v in vsource if v in vtarget)
                self.variables = [v for v in vlist]

            elif source_is_topo:
                # source = topo, target = operator
                vtarget = self._target.variables
                if not has_var:
                    vlist = (v for v in vtarget)
                self.variables = [v for v in vlist]

            else:
                # source = operator, target = topo
                vsource = self._source.variables
                if not has_var:
                    vlist = (v for v in vsource)
                self.variables = [v for v in vlist]

        assert len(self.variables) > 0

        # Variables is converted to a dict to be coherent with
        # computational operators ...
        self.variables = {key: None for key in self.variables}

        # All variables must have the same domain
        self.domain = self.variables.keys()[0].domain
        for v in self.variables:
            assert v.domain is self.domain

    def _set_topology(self, current):
        """This function check if current is valid, fits with self.variables
        and get its topology to set self._topology.

        Parameters
        ----------
        current : :class:`~hysop.topology.topology.CartesianTopology` or
         :class:`~hysop.core.mpi.operator.computational.Computational`

        """
        if isinstance(current, CartesianTopology):
            result = current
            for v in self.variables:
                v.discretize(result)
        elif isinstance(current, Computational):
            self._check_operator(current)
            vref = self.variables.keys()[0]
            vcurrent = current.variables
            result = vcurrent[vref]
            # We ensure that all vars have
            # the same topo in target/target.
            for v in (v for v in self.variables if v is not vref):
                assert vcurrent[v] is result
        else:
            msg = "the source/target is neither an operator or a topology."
            raise AttributeError(msg)
        assert result.task_id() == self.domain.current_task()
        return result

    def computation_time(self):
        pass


class RedistributeIntra(Redistribute):
    """Data transfer between two operators/topologies.
    Source and target must:
    - be defined on the same communicator
    - work on the same number of mpi process
    - work with the same global resolution
    """

    def __init__(self, **kwds):
        """Data transfer between two operators/topologies defined on the
        same communicator

        Source and target must:
        * be defined on the same communicator
        * work on the same number of mpi process
        * work with the same global resolution
        """

        # Base class initialisation
        super(RedistributeIntra, self).__init__(**kwds)

        # Warning : comm from io_params will be used as
        # reference for all mpi communication of this operator.
        # --> rank computed in refcomm
        # --> source and target must work inside refcomm
        # If io_params is None, refcomm will COMM_WORLD.

        # Dictionnary of discrete fields to be sent
        self._vsource = {}
        # Dictionnary of discrete fields to be overwritten
        self._vtarget = {}

        # dictionnary which maps rank with mpi derived type
        # for send operations
        self._send = {}
        # dictionnay which maps rank with mpi derived type
        # for send operations
        self._receive = {}
        # dictionnary which map rank/field name with a
        # receive request
        self._r_request = None
        # dictionnary which map rank/field name with a
        # send request
        self._s_request = None

        # Set list of variables and the domain.
        self._set_variables()
        # Set mpi related stuff
        self._set_domain_and_tasks()

    @opsetup
    def setup(self, rwork=None, iwork=None):
        # At setup, source and topo must be either
        # a hysop.topology.topology.CartesianTopology or
        # a computational operator.

        msg = 'Redistribute error : undefined source of target.'
        assert self._source is not None and self._target is not None, msg

        t_source = self._set_topology(self._source)
        t_target = self._set_topology(self._target)

        source_res = t_source.mesh.discretization.resolution
        target_res = t_target.mesh.discretization.resolution
        msg = 'Redistribute error: source and target must '
        msg += 'have the same global resolution.'
        assert (source_res == target_res).all(), msg

        # Set the dictionnaries of source/target variables
        self._vsource = {v: v.discretize(t_source)
                         for v in self.variables}
        self._vtarget = {v: v.discretize(t_target)
                         for v in self.variables}

        # We can create the bridge
        self.bridge = Bridge(t_source, t_target)

        # Shape of reference is the shape of source/target mesh
        self._send = self.bridge.send_types()
        self._receive = self.bridge.recv_types()
        self._set_synchro()
        self._is_uptodate = True

    def _set_synchro(self):
        """
        Set who must wait for who ...
        """
        # Check input operators
        if isinstance(self._source, Computational):
            #  redistribute must wait for source if a variable of redistribute
            # is an output from source.
            for v in self.variables:
                vout = v in self._source.output or False
            if vout:
                self.wait_for(self._source)
                # And source must wait for redistribute
                # if a variable of red. is an output from source.
                self._source.wait_for(self)

        if isinstance(self._target, Computational):
            # target operator must wait for
            # the end of this operator to apply.
            self._run_till.append(self._target)

        # Add this operator into wait list of
        # operators listed in run_till
        for op in self._run_till:
            op.wait_for(self)

        self._is_uptodate = True

    def add_run_till_op(self, op):
        """Add an operator to the wait list"""
        self._run_till.append(op)
        op.wait_for(self)

    @opapply
    def apply(self, simulation=None):
        # Try different way to send vars?
        # - Buffered : copy all data into a buffer and send/recv
        # - Standard : one send/recv per component
        # --- Standard send/recv ---
        br = self.bridge

        # reset send/recv requests
        self._r_request = {}
        self._s_request = {}

        basetag = self.mpi_params.rank + 1
        # Comm used for send/receive operations
        # It must contains all proc. of source topo and
        # target topo.
        refcomm = self.bridge.comm
        # Loop over all required components of each variable
        for v in self.variables:
            for d in self._range_components(v):
                v_name = v.name + DirectionLabels[d]

                # Deal with local copies of data
                if br.has_local_inter():
                    vTo = self._vtarget[v].data[d]
                    vFrom = self._vsource[v].data[d]
                    vTo[br.local_target_ind()] = vFrom[br.local_source_ind()]

                # Transfers to other mpi processes
                for rk in self._receive:
                    recvtag = basetag * 989 + (rk + 1) * 99 + (d + 1) * 88
                    mpi_type = self._receive[rk]
                    vTo = self._vtarget[v].data[d]
                    self._r_request[v_name + str(rk)] = \
                        refcomm.Irecv([vTo, 1, mpi_type],
                                      source=rk, tag=recvtag)
                    self._has_requests = True
                for rk in self._send:
                    sendtag = (rk + 1) * 989 + basetag * 99 + (d + 1) * 88
                    mpi_type = self._send[rk]
                    vFrom = self._vsource[v].data[d]
                    self._s_request[v_name + str(rk)] = \
                        refcomm.Issend([vFrom, 1, mpi_type],
                                       dest=rk, tag=sendtag)
                    self._has_requests = True

    def wait(self):
        if self._has_requests:
            for rk in self._r_request:
                self._r_request[rk].Wait()
            for rk in self._s_request:
                self._s_request[rk].Wait()
        self._has_requests = False

    def test_requests(self):
        res = True
        for rk in self._r_request.keys():
            res = self._r_request[rk].Test()
            if not res:
                return res
        for rk in self._s_request.keys():
            res = self._s_request[rk].Test()
            if not res:
                return res

    def test_single_request(self, rsend=None, rrecv=None):
        """if neither rsend or rrecv is given return
        True if all communication request are complete
        else check for sending to rsend or receiving from rrecv.
        Process ranks should be those in parent_comm.

        Parameters
        ----------
        rsend : string
            discrete variable name + DirectionLabels + rank of the process
            to which a message has been sent
            and for which we want to test message completion.
        rrecv : string
            discrete variable name + DirectionLabels + rank of the process
            from which a message has been receive
            and for which we want to test message completion.

        """
        if rsend is not None or rrecv is not None:
            send_res = True
            recv_res = True
            if rsend is not None:
                send_res = self._s_request[rsend].Test()
            if rrecv is not None:
                recv_res = self._r_request[rrecv].Test()
            res = send_res and recv_res
            return res
        else:
            return self.test_requests()


class RedistributeInter(Redistribute):
    """Operator to redistribute data from one communicator to another.
    Source/target may be either a topology or a computational operator.
    It implies mpi inter-communications.
    """

    @debug
    def __init__(self, parent, source_id=None, target_id=None, **kwds):
        """redistribute data from one communicator to another.
        Source/target may be either a topology or a computational operator.
        It implies mpi inter-communications.

        Parameters
        ----------
        parent : MPI.COMM
            mpi communicator that must owns all the
            processes involved in source and target.
        source_id, target_id : int
            mpi task ids for the source/target.
            Required if source/target is None
            else infered from source/target.

        See other required parameters in base class.
        """
        super(RedistributeInter, self).__init__(**kwds)

        # parent communicator, that must contains all processes
        # involved in source and target tasks.
        self.parent = parent

        # set source and targets ids.
        # They must be known before setup.
        # Either they can be infered from source and target
        # or must be set in argument list, if either source
        # or target is undefined on the current process.
        if self._source is None:
            assert source_id is not None

        if self._target is None:
            assert target_id is not None

        self._source_id = source_id
        self._target_id = target_id

        # Set list of variables and domain.
        self._set_variables()
        # Set mpi related stuff
        self._set_domain_and_tasks()

        # Domain is set, we can check if we are on source or target
        current_task = self.domain.current_task()
        self._is_source = current_task == self._source_id
        self._is_target = current_task == self._target_id
        assert self._is_target or self._is_source
        assert not (self._is_target and self._is_source)

        nbprocs = len(self.domain.tasks_list())
        msg = "Parent communicator size and number of procs "
        msg += "in domain differ."
        assert parent.Get_size() == nbprocs, msg

        # the local topology. May be either source or target
        # depending on the task of the current process.
        self._topology = None

        # dictionnary which maps rank with mpi derived type
        # used for send/recv operations (send on source, recv on target ...)
        self._transfer_types = None

        # dictionnary which maps rank/field name with a
        # send/recv request
        self._requests = {}

    @debug
    @opsetup
    def setup(self, rwork=None, iwork=None):
        # First of all, we need to get the current topology:
        if self._is_source:
            assert self._source is not None
            self._topology = self._set_topology(self._source)
        elif self._is_target:
            assert self._target is not None
            self._topology = self._set_topology(self._target)

        # Now we can build the bridge (intercomm)
        self.bridge = BridgeInter(self._topology, self.parent,
                                  self._source_id, self._target_id)

        # And get mpi derived types
        self._transfer_types = self.bridge.transfer_types()

        self._set_synchro()
        self._is_uptodate = True

    def _set_synchro(self):
        """
        Set who must wait for who ...
        """
        if self._is_source and isinstance(self._source, Computational):
            #  redistribute must wait for source if a variable of redistribute
            # is an output from source.
            for v in self.variables:
                vout = v in self._source.output or False
            if vout:
                self.wait_for(self._source)
                # And source must wait for redistribute
                # if a variable of red. is an output from source.
                self._source.wait_for(self)

        if self._is_target and isinstance(self._target, Computational):
            # target operator must wait for
            # the end of this operator to apply.
            self._run_till.append(self._target)

        # Add this operator into wait list of
        # operators listed in run_till
        for op in self._run_till:
            op.wait_for(self)

    def add_run_till_op(self, op):
        """Add an operator to the wait list"""
        if self._is_target:
            self._run_till.append(op)
            op.wait_for(self)

    @debug
    @opapply
    def apply(self, simulation=None):
        # --- Standard send/recv ---
        self._requests = {}

        # basetag = self.mpi_params.rank + 1
        # Comm used for send/receive operations
        # It must contains all proc. of source topo and
        # target topo.
        refcomm = self.bridge.comm
        # Map between rank and mpi types
        # Loop over all required components of each variable
        for v in self.variables:
            rank = self._topology.comm.Get_rank()
            for d in self._range_components(v):
                v_name = v.name + DirectionLabels[d]
                vtab = v.discrete_fields[self._topology].data[d]
                for rk in self._transfer_types:
                    if self._is_target:
                        # Set reception
                        self._requests[v_name + str(rk)] = \
                            refcomm.Irecv([vtab[...], 1,
                                           self._transfer_types[rk]],
                                          source=rk, tag=rk)
                    if self._is_source:
                        self._requests[v_name + str(rk)] = \
                            refcomm.Issend([vtab[...], 1,
                                            self._transfer_types[rk]],
                                           dest=rk, tag=rank)
                    self._has_requests = True

    def wait(self):
        if self._has_requests:
            for rk in self._requests:
                self._requests[rk].Wait()
        for v in self.variables:
            for d in self._range_components(v):
                vtab = v.discrete_fields[self._topology].data[d]
        self._has_requests = False

    def test_requests(self):
        res = True
        for rk in self._requests:
            res = self._requests[rk].Test()
            if not res:
                return res


class RedistributeOverlap(RedistributeIntra):
    """A specific redistribute where source and target do not work with the same
    group of mpi processes.
    Requirements :
    - work only on topologies, not on operators
    - same global resolution for both topologies
    - group from source topology and target topology MUST overlap.
    """

    @opsetup
    def setup(self, rwork=None, iwork=None):
        """
        Check/set the list of variables to be distributed

        What must be set at setup?
        ---> the list of continuous variables to be distributed
        ---> the bridge (one for all variables, which means
        that all vars must have the same topology in source
        and the same topology in target.
        ---> the list of discrete variables for source and
        for target.
        """
        if self._source is not None:
            self._vsource = self._discrete_fields(self._source)
        if self._target is not None:
            self._vtarget = self._discrete_fields(self._target)

        # We can create the bridge
        self.bridge = BridgeOverlap(source=self._source, target=self._target,
                                    comm_ref=self.mpi_params.comm)

        # Build mpi derived types for send and receive operations.
        # Shape of reference is the shape of source/target mesh
        if self._source is not None:
            self._send = self.bridge.send_types()
        if self._target is not None:
            self._receive = self.bridge.recv_types()

        self._set_synchro()
        self._is_uptodate = True

    def _discrete_fields(self, topo):
        """Return the dictionnary of discrete fields for topo
        and the variables of this operator.

        Parameters
        ----------
        topo : :class:`~hysop.topology.topology.CartesianTopology`
        """
        assert isinstance(topo, CartesianTopology)
        return {v: v.discretize(topo) for v in self.variables}
