"""Testing Scales advection operator.
"""
from __future__ import print_function
import numpy as np
from hysop.methods import TimeIntegrator, Interpolation,\
    Remesh, Support, Splitting
from hysop.methods import RK2, L2_1, L4_2, M8Prime, Linear
from hysop.domain.box import Box
from hysop.fields.continuous_field import Field
from hysop.operator.advection import Advection, ScalesAdvection
from hysop.problem.simulation import O2FULLHALF
from hysop.problem.simulation import Simulation
from hysop.testsenv import scales_failed
from hysop.tools.parameters import Discretization
d3d = Discretization([17, 17, 17])
d3d_g = Discretization([17, 17, 17], [6, 6, 6])

m1 = {TimeIntegrator: RK2, Interpolation: Linear, Remesh: L2_1,
      Splitting: O2FULLHALF, Support: ''}
m2 = {TimeIntegrator: RK2, Interpolation: Linear, Remesh: L4_2,
      Splitting: O2FULLHALF, Support: ''}
m3 = {TimeIntegrator: RK2, Interpolation: Linear, Remesh: M8Prime,
      Splitting: O2FULLHALF, Support: ''}


def run_advection(vector_field, method, mref=None, random_velocity=False):
    """Create advection operator, ref operator
    and fields, run scales and python advection,
    compare results.

    Parameters
    ----------
    vector_field: bool
        True to advect a vector field, else scalar field.
    method : dictionnary
        Set scales remeshing type.
    mref : dictionnary, optional
        method for pure python advection. Default = m1
    random_velocity: boolean
        If True, randomize velocity values, else set it to 0.
    """
    box = Box(length=[1., 1., 1.], origin=[0., 0., 0.])
    scal = Field(domain=box, name='Scalar', is_vector=vector_field)
    scal_ref = Field(domain=box, name='Scalar_ref', is_vector=vector_field)
    velo = Field(domain=box, name='Velocity',
                 formula=lambda x, y, z, t: (0., 0., 0.), is_vector=True,
                 vectorize_formula=True)
    advec = ScalesAdvection(velocity=velo,
                            advected_fields=scal,
                            discretization=d3d,
                            method=method)
    if mref is None:
        mref = m1
    advec_py = Advection(velocity=velo,
                         advected_fields=scal_ref, discretization=d3d_g,
                         method=mref)
    advec.discretize()
    advec_py.discretize()
    advec.setup()
    advec_py.setup()
    # Get and randomize discrete fields
    topo = advec.advected_fields_topology()
    topo_ref = advec_py.advected_fields_topology()

    if random_velocity:
        topo_velo = advec.velocity_topology()
        vd = velo.randomize(topo_velo)
        for d in xrange(velo.nb_components):
            vd[d] /= 2 * topo_velo.mesh.local_resolution[d]
    else:
        assert (velo.norm(topo) == 0).all()
    ic = topo.mesh.compute_index
    icref = topo_ref.mesh.compute_index
    print(icref)
    print(ic)
    scal_d = scal.randomize(topo)
    scal_ref_d = scal.discretize(topo_ref)
    for d in xrange(len(scal_d.data)):
        scal_ref_d[d][icref] = scal_d[d][ic]
        assert np.allclose(scal_ref_d.data[d][icref], scal_d.data[d][ic])
    advec.apply(Simulation(start=0., end=0.1, nb_iter=1))
    advec_py.apply(Simulation(start=0., end=0.1, nb_iter=1))
    # compare values on grid (excluding ghosts)
    # for d in xrange(len(scal_d.data)):
    #     assert np.allclose(scal_ref_d.data[d][icref], scal_d.data[d][ic])


def run_advection_2(vector_field, method):
    """Create advection operator,
    fields, run and check results.

    Parameters
    ----------
    vector_field: bool
        True to advect a vector field, else scalar field.
    method : dictionnary
        Set scales remeshing type.
    """
    box = Box(length=[1., 1., 1.], origin=[0., 0., 0.])
    scal = Field(domain=box, name='Scalar', is_vector=vector_field)
    scal_ref = Field(domain=box, name='Scalar_ref', is_vector=vector_field)
    velo = Field(domain=box, name='Velocity',
                 formula=lambda x, y, z, t: (0., 0., 0.), is_vector=True,
                 vectorize_formula=True)
    advec = ScalesAdvection(velocity=velo,
                            advected_fields=scal, discretization=d3d,
                            method=method)
    advec.discretize()
    advec.setup()
    # Get and randomize discrete fields
    topo = advec.advected_fields_topology()
    assert (velo.norm(topo) == 0).all()
    ic = topo.mesh.compute_index
    scal_d = scal.randomize(topo)
    scal_ref.copy(scal, topo)
    scal_ref_d = scal.discretize(topo)
    for d in xrange(len(scal_d.data)):
        assert np.allclose(scal_ref_d.data[d][ic], scal_d.data[d][ic])
    advec.apply(Simulation(start=0., end=0.1, nb_iter=1))
    # compare values on grid (excluding ghosts)
    for d in xrange(len(scal_d.data)):
        assert np.allclose(scal_ref_d.data[d][ic], scal_d.data[d][ic])


@scales_failed
def test_null_velocity_m4():
    """Scalar field advection, null velocity, M4prime remesh
    """
    method = {Remesh: 'p_M4'}
    run_advection_2(False, method)


@scales_failed
def test_null_velocity_vec_m4():
    """Vector field advection, null velocity, M4prime remesh
    """
    method = {Remesh: 'p_M4'}
    run_advection_2(True, method)


@scales_failed
def test_null_velocity_m6():
    """Scalar field advection, null velocity, M6prime remesh
    """
    method = {Remesh: 'p_M6'}
    run_advection_2(False, method)


@scales_failed
def test_null_velocity_vec_m6():
    """Vector field advection, null velocity, M6prime remesh
    """
    method = {Remesh: 'p_M6'}
    run_advection_2(True, method)


@scales_failed
def test_null_velocity_m8():
    """Scalar field advection, null velocity, M6prime remesh
    """
    method = {Remesh: 'p_M8'}
    run_advection_2(False, method)


@scales_failed
def test_null_velocity_vec_m8():
    """Vector field advection, null velocity, M6prime remesh
    """
    method = {Remesh: 'p_M8'}
    run_advection_2(True, method)


@scales_failed
def test_random_velocity_m4():
    """Scalar field advection, random velocity, M4prime remesh
    """
    method = {Remesh: 'p_M4'}
    run_advection(False, method, random_velocity=True)


@scales_failed
def test_random_velocity_vec_m4():
    """Vector field advection, random velocity, M4prime remesh
    """
    method = {Remesh: 'p_M4'}
    run_advection(True, method, random_velocity=True)


@scales_failed
def test_random_velocity_m6():
    """Scalar field advection, random velocity, M6prime remesh
    """
    method = {Remesh: 'p_M6'}
    run_advection(False, method, m2, random_velocity=True)


@scales_failed
def test_random_velocity_vec_m6():
    """Vector field advection, random velocity, M6prime remesh
    """
    method = {Remesh: 'p_M6'}
    run_advection(True, method, m2, random_velocity=True)


@scales_failed
def test_random_velocity_m8():
    """Scalar field advection, random velocity, M8prime remesh
    """
    method = {Remesh: 'p_M8'}
    run_advection(False, method, m3, random_velocity=True)


@scales_failed
def test_random_velocity_vec_m8():
    """Vector field advection, random velocity, M8prime remesh
    """
    method = {Remesh: 'p_M8'}
    run_advection(True, method, m3, random_velocity=True)
