# -*- coding: utf-8 -*-
"""Operator used to shift velocity value
 to fit with a required input flowrate.

Check details in :ref:`velocity_correction` in HySoP user guide.


"""
from hysop.constants import debug
from hysop.operator.discrete.velocity_correction import VelocityCorrection_D
from hysop.operator.computational import Computational
from hysop.domain.control_box import ControlBox
from hysop.operator.continuous import opsetup


class VelocityCorrection(Computational):
    """The velocity field is corrected after solving the
    Poisson equation. For more details about calculations,
    see :ref:`velocity_correction` in HySoP user guide.
    """

    @debug
    def __init__(self, velocity, vorticity, req_flowrate, **kwds):
        """Update velocity field (solution of Poisson equation)
        in order to prescribe proper mean flow and ensure
        the desired inlet flowrate.

        Parameters
        ----------
        velocity : :class:`~hysop.fields.continuous_field.Field`
            in/out velocity continuous vector field.
        vorticity : :class:`~hysop.fields.continuous_field.Field`
            input vorticity vector field.
        req_flowrate : a
          :class:`~hysop.fields.variable_parameter.VariableParameter`
            the desired inlet flowrate value
        kwds : base class parameters
        """
        assert 'variables' not in kwds, 'variables parameter is useless.'
        super(VelocityCorrection, self).__init__(
            variables=[velocity, vorticity], **kwds)
        # velocity field
        self.velocity = velocity
        # vorticity field
        self.vorticity = vorticity
        self.input = [self.velocity, self.vorticity]
        self.output = [self.velocity]
        # Expected value for the flow rate through input surface
        self.req_flowrate = req_flowrate
        dom = self.velocity.domain
        # volume of control used to compute a reference for correction.
        self.cb = ControlBox(origin=dom.origin, length=dom.length,
                             parent=dom)

    def discretize(self):
        super(VelocityCorrection, self)._standard_discretize()
        assert self._single_topo, 'Multi-resolution case is not allowed.'

    @debug
    @opsetup
    def setup(self, rwork=None, iwork=None):
        if not self._is_uptodate:
            self.discrete_op =\
                VelocityCorrection_D(self.discrete_fields[self.velocity],
                                     self.discrete_fields[self.vorticity],
                                     self.req_flowrate, self.cb, rwork=rwork,
                                     iwork=iwork)
            # Output setup
            self._set_io('velocity_correction', (1, 2 + self.domain.dim))
            self.discrete_op.set_writer(self._writer)
            self._is_uptodate = True

    def compute_correction(self):
        """Compute the required correction for the current state
        but do not apply it onto velocity.
        """
        self.discrete_op.compute_correction()

    def get_work_properties(self):
        return {'rwork': None, 'iwork': None}
