#!/usr/bin/env python

import hysop

from hysop.methods_keys import TimeIntegrator, Interpolation, Remesh,\
    Support, Splitting, MultiScale, MultiScale, SpaceDiscretisation, \
    GhostUpdate, Scales, dtCrit, ExtraArgs

from hysop.problem.simulation        import Simulation
from hysop.fields.variable_parameter import VariableParameter

from hysop.numerics.odesolvers import RK2 as RK2
from hysop.numerics.odesolvers import RK3 as RK3
from hysop.numerics.odesolvers import RK4 as RK4

from hysop.numerics.finite_differences import FD_C_4
from hysop.numerics.remeshing import L6_4 as remesh_formula
from hysop.numerics.remeshing import Linear

from hysop.operator.advection import Advection
from hysop.operator.diffusion import Diffusion
from hysop.operator.stretching import Stretching
from hysop.operator.poisson import Poisson
from hysop.operator.adapt_timestep import AdaptTimeStep

from hysop.operator.redistribute_inter import RedistributeInter
from hysop.operator.redistribute_intra import RedistributeIntra

from hysop.gpu.gpu_transfer import DataTransfer
from hysop.operator.hdf_io  import HDF_Writer

import utilities as utils


## Simulation parameters
dim = 3
N   = 1<<8 + 1
nb_elems = [N]*dim

## Physical parameters:
# Flow viscosity
viscosity = 1e-4

## I/O parameters
dt_output    = IOParams(frequency=out_freq, filename='cpu_dt.dat',    fileformat=ASCII)
velo_output  = IOParams(frequency=out_freq, filename='cpu_velo.dat',  fileformat=HDF5)
vorti_output = IOParams(frequency=out_freq, filename='cpu_vorti.dat', fileformat=HDF5)


# Domain
box = hysop.Box(length=[1.0]*dim, 
                origin=[0.0]*dim,
                bc=PERIODIC)

# Fields
velo  = hysop.Field(domain=box, formula=utils.initialize_velocity,
                   name='Velocity',  is_vector=True)
vorti = hysop.Field(domain=box,
                   name='Vorticity', is_vector=True)

# Discretizations
d_uw = Discretization(nb_elems)

# Topologies
topo_uw = box.create_topology(d_uw, dim=1)

# Adaptative timestep simulation
dt   = VariableParameter({'dt': 0.001})
simu = Simulation(start=0.0, end=5.0, time_step=0.001, max_iter=10000)


## Operators
ops = {}

# CPU operators
ops['advec_w'] = Advection(velo,
                     discretization=topo_uw,
                     variables={vorti: topo_uw},
                     method={Scales: 'p_64', MultiScale: 'L4_4'})
ops['stretching'] = Stretching(velo, vorti, discretization=d_uw)
ops['diffusion']  = Diffusion(variables={vorti: d_uw}, viscosity=viscosity)
ops['poisson']    = Poisson(velo, vorti, discretization=d_uw)


# other operators
ops['dt_adapt'] = AdaptTimeStep(velo, vorti,
                         simulation=simu,
                         time_range=[0, np.infty],
                         discretization=d_uw,
                         method={TimeIntegrator: RK3,
                                 SpaceDiscretisation: FD_C_4,
                                 dtCrit: ['gradU', 'cfl']},
                         lcfl=0.15,
                         cfl=1.5,
                         io_params=dt_output)

ops['velo_io']  = HDF_Writer(variables={velo : topo_uw}, io_params=velo_output)
ops['vorti_io'] = HDF_Writer(variables={vorti: topo_uw}, io_params=vorti_output)

for op in ops:
    op.discretize()


